#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import AnyField, StringField, KeyField
from ...lib.sparse.core import get_sparse_module, get_array_module, naked
from ...lib.sparse import SparseNDArray
from ..array_utils import create_array
from ..utils import get_order
from .core import TensorNoInput, TensorLike
from .array import tensor


class TensorFull(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_FULL

    _fill_value = AnyField('fill_value')
    _order = StringField('order')

    def __init__(self, fill_value=None, dtype=None, gpu=None, order=None, **kw):
        if dtype is not None:
            dtype = np.dtype(dtype)
            if fill_value is not None:
                fill_value = dtype.type(fill_value)
        elif fill_value is not None:
            dtype = np.array(fill_value).dtype
        super().__init__(_fill_value=fill_value, _dtype=dtype, _gpu=gpu, _order=order, **kw)

    @property
    def fill_value(self):
        return self._fill_value

    @property
    def order(self):
        return self._order

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        ctx[chunk.key] = create_array(op)('full', chunk.shape,
                                          op.fill_value, dtype=op.dtype, order=op.order)


def full(shape, fill_value, dtype=None, chunk_size=None, gpu=False, order='C'):
    """
    Return a new tensor of given shape and type, filled with `fill_value`.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the new tensor, e.g., ``(2, 3)`` or ``2``.
    fill_value : scalar
        Fill value.
    dtype : data-type, optional
        The desired data-type for the tensor  The default, `None`, means
         `np.array(fill_value).dtype`.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    order : {'C', 'F'}, optional
        Whether to store multidimensional data in C- or Fortran-contiguous
        (row- or column-wise) order in memory.

    Returns
    -------
    out : Tensor
        Tensor of `fill_value` with the given shape, dtype, and order.

    See Also
    --------
    zeros_like : Return a tensor of zeros with shape and type of input.
    ones_like : Return a tensor of ones with shape and type of input.
    empty_like : Return an empty tensor with shape and type of input.
    full_like : Fill a tensor with shape and type of input.
    zeros : Return a new tensor setting values to zero.
    ones : Return a new tensor setting values to one.
    empty : Return a new uninitialized tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.full((2, 2), mt.inf).execute()
    array([[ inf,  inf],
           [ inf,  inf]])
    >>> mt.full((2, 2), 10).execute()
    array([[10, 10],
           [10, 10]])

    """
    v = np.asarray(fill_value)
    if len(v.shape) > 0:
        from ..base import broadcast_to
        return broadcast_to(
            tensor(v, dtype=dtype, chunk_size=chunk_size, gpu=gpu, order=order), shape)

    tensor_order = get_order(order, None, available_options='CF',
                             err_msg="only 'C' or 'F' order is permitted")
    op = TensorFull(fill_value, dtype=dtype, gpu=gpu, order=order)
    return op(shape, chunk_size=chunk_size, order=tensor_order)


class TensorFullLike(TensorLike):
    _op_type_ = OperandDef.TENSOR_FULL_LIKE

    _input = KeyField('input')
    _fill_value = AnyField('fill_value')
    _order = StringField('order')

    def __init__(self, fill_value=None, dtype=None, gpu=None, sparse=False, order=None, **kw):
        if dtype is not None:
            dtype = np.dtype(dtype)
            if fill_value is not None:
                fill_value = dtype.type(fill_value)
        elif fill_value is not None:
            dtype = np.array(fill_value).dtype
        super().__init__(_fill_value=fill_value, _dtype=dtype, _gpu=gpu, _order=order,
                         _sparse=sparse, **kw)

    @property
    def fill_value(self):
        return self._fill_value

    @property
    def order(self):
        return self._order

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        if op.issparse():
            in_data = naked(ctx[op.inputs[0].key])
            xps = get_sparse_module(in_data)
            xp = get_array_module(in_data)
            ctx[chunk.key] = SparseNDArray(xps.csr_matrix(
                (xp.full_like(in_data.data, op.fill_value, dtype=op.dtype),
                 in_data.indices, in_data.indptr), shape=in_data.shape
            ))
        else:
            ctx[chunk.key] = create_array(op)(
                'full_like', ctx[op.inputs[0].key], op.fill_value,
                dtype=op.dtype, order=op.order)


def full_like(a, fill_value, dtype=None, gpu=None, order='K'):
    """
    Return a full tensor with the same shape and type as a given tensor.

    Parameters
    ----------
    a : array_like
        The shape and data-type of `a` define these same attributes of
        the returned tensor.
    fill_value : scalar
        Fill value.
    dtype : data-type, optional
        Overrides the data type of the result.
    gpu : bool, optional
        Allocate the tensor on GPU if True, None as default
    order : {'C', 'F', 'A', or 'K'}, optional
        Overrides the memory layout of the result. 'C' means C-order,
        'F' means F-order, 'A' means 'F' if `a` is Fortran contiguous,
        'C' otherwise. 'K' means match the layout of `a` as closely
        as possible.

    Returns
    -------
    out : Tensor
        Tensor of `fill_value` with the same shape and type as `a`.

    See Also
    --------
    empty_like : Return an empty tensor with shape and type of input.
    ones_like : Return a tensor of ones with shape and type of input.
    zeros_like : Return a tensor of zeros with shape and type of input.
    full : Return a new tensor of given shape filled with value.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> x = mt.arange(6, dtype=int)
    >>> mt.full_like(x, 1).execute()
    array([1, 1, 1, 1, 1, 1])
    >>> mt.full_like(x, 0.1).execute()
    array([0, 0, 0, 0, 0, 0])
    >>> mt.full_like(x, 0.1, dtype=mt.double).execute()
    array([ 0.1,  0.1,  0.1,  0.1,  0.1,  0.1])
    >>> mt.full_like(x, mt.nan, dtype=mt.double).execute()
    array([ nan,  nan,  nan,  nan,  nan,  nan])

    >>> y = mt.arange(6, dtype=mt.double)
    >>> mt.full_like(y, 0.1).execute()
    array([ 0.1,  0.1,  0.1,  0.1,  0.1,  0.1])

    """
    a = tensor(a)
    tensor_order = get_order(order, a.order)
    if dtype is None:
        dtype = a.dtype
    gpu = a.op.gpu if gpu is None else gpu
    op = TensorFullLike(fill_value=fill_value, dtype=dtype,
                        gpu=gpu, sparse=a.issparse())
    return op(a, order=tensor_order)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import tempfile
import subprocess

import numpy as np


from .. import opcodes
from ..operands import MergeDictOperand, OutputType
from ..context import RunningMode
from ..serialize import BytesField, ListField, Int32Field, StringField, BoolField
from ..utils import to_binary


class RunScript(MergeDictOperand):
    _op_type_ = opcodes.RUN_SCRIPT

    _code = BytesField('code')
    _mode = StringField('mode')
    _retry_when_fail = BoolField('retry_when_fail')
    _command_args = ListField('command_args')
    _world_size = Int32Field('world_size')
    _rank = Int32Field('rank')

    def __init__(self, code=None, mode=None, world_size=None, rank=None, retry_when_fail=None,
                 command_args=None, **kw):
        super().__init__(_code=code, _mode=mode, _world_size=world_size, _rank=rank,
                         _retry_when_fail=retry_when_fail, _command_args=command_args,
                         **kw)
        if self.output_types is None:
            self.output_types = [OutputType.object]

    @property
    def code(self):
        return self._code

    @property
    def mode(self):
        return self._mode

    @property
    def world_size(self):
        return self._world_size

    @property
    def rank(self):
        return self._rank

    @property
    def command_args(self):
        return self._command_args or []

    @property
    def retryable(self):
        return self._retry_when_fail

    def __call__(self):
        return self.new_tileable(None)

    @classmethod
    def tile(cls, op):
        out_chunks = []
        for i in range(op.world_size):
            chunk_op = op.copy().reset_key()
            chunk_op._rank = i
            out_chunks.append(chunk_op.new_chunk(None, index=(i,)))

        new_op = op.copy()
        return new_op.new_tileables(op.inputs, chunks=out_chunks,
                                    nsplits=(tuple(np.nan for _ in range(len(out_chunks))),))

    @classmethod
    def _execute_with_subprocess(cls, op, envs=None):
        # write source code into a temp file
        fd, filename = tempfile.mkstemp('.py')
        with os.fdopen(fd, 'wb') as f:
            f.write(op.code)

        new_envs = os.environ.copy()
        new_envs.update(envs or dict())
        try:
            # exec code in a new process
            process = subprocess.Popen([sys.executable, filename] + op.command_args,
                                       env=new_envs)
            process.wait()
            if process.returncode != 0:  # pragma: no cover
                raise RuntimeError('Run script failed')

        finally:
            os.remove(filename)

    @classmethod
    def _execute_with_exec(cls, op, local=None):
        local = local or dict()

        try:
            exec(op.code, local)
        finally:
            sys.stdout.flush()

    @classmethod
    def _build_envs(cls, ctx, op):
        # set mars envs
        envs = dict()
        envs['MARS_SESSION_ID'] = str(ctx.session_id)
        envs['RANK'] = str(op.rank)
        envs['WORLD_SIZE'] = str(op.world_size)
        if ctx.running_mode != RunningMode.local:
            envs['MARS_SCHEDULER_ADDRESS'] = str(ctx._scheduler_address)
        return envs

    @classmethod
    def _build_locals(cls, ctx, op):
        sess = ctx.get_current_session().as_default()

        return dict(session=sess)

    @classmethod
    def execute(cls, ctx, op):
        if op.merge:
            return super().execute(ctx, op)

        old_env = os.environ.copy()
        envs = cls._build_envs(ctx, op)

        try:
            if op.mode == 'spawn':
                cls._execute_with_subprocess(op, envs)
            elif op.mode == 'exec':
                os.environ.update(envs)
                cls._execute_with_exec(op, cls._build_locals(ctx, op))
            else:  # pragma: no cover
                raise TypeError(f'Unsupported mode {op.mode}')

            if op.rank == 0:
                ctx[op.outputs[0].key] = {'status': 'ok'}
            else:
                ctx[op.outputs[0].key] = {}
        finally:
            os.environ = old_env


def run_script(script, n_workers=1, mode='spawn', command_argv=None,
               session=None, retry_when_fail=False, run_kwargs=None):
    """
    Run script in Mars cluster.

    Parameters
    ----------
    script: str or file-like object
        Script to run.
    n_workers: int
        number of workers to run the script
    mode: str, 'spawn' as default
        'spawn' mode use subprocess to run script,
        'exec' mode use exec to run in current process.
    command_argv: list
        extra command args for script
    session: Mars session
        if not provided, will use default one
    retry_when_fail: bool, default False
       If True, retry when function failed.
    run_kwargs: dict
        extra kwargs for session.run

    Returns
    -------
    Object
        Mars Object.

    """

    if hasattr(script, 'read'):
        code = script.read()
    else:
        with open(os.path.abspath(script), 'rb') as f:
            code = f.read()
    if mode not in ['exec', 'spawn']:  # pragma: no cover
        raise TypeError(f'Unsupported mode {mode}')

    op = RunScript(code=to_binary(code), mode=mode, world_size=n_workers,
                   retry_when_fail=retry_when_fail, command_args=command_argv)
    return op().execute(session=session, **(run_kwargs or {})).fetch(session=session)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def _install():
    from .to_csv import to_csv
    from .to_sql import to_sql
    from .to_vineyard import to_vineyard
    from ..operands import DATAFRAME_TYPE, SERIES_TYPE

    for cls in DATAFRAME_TYPE:
        setattr(cls, 'to_csv', to_csv)
        setattr(cls, 'to_sql', to_sql)
        setattr(cls, 'to_vineyard', to_vineyard)

    for cls in SERIES_TYPE:
        setattr(cls, 'to_csv', to_csv)
        setattr(cls, 'to_sql', to_sql)


_install()
del _install



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


def wait_services_ready(selectors, min_counts, count_fun, timeout=None):
    readies = [0] * len(selectors)
    start_time = time.time()
    while True:
        all_satisfy = True
        for idx, selector in enumerate(selectors):
            if readies[idx] < min_counts[idx]:
                all_satisfy = False
                readies[idx] = count_fun(selector)
                break
        if all_satisfy:
            break
        if timeout and timeout + start_time < time.time():
            raise TimeoutError('Wait kubernetes cluster start timeout')
        time.sleep(1)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle

import numpy as np
import pandas as pd

from .... import opcodes as OperandDef
from ....serialize import KeyField, BytesField
from ....dataframe.core import SERIES_CHUNK_TYPE, DATAFRAME_CHUNK_TYPE
from ....dataframe.utils import parse_index
from ....tensor.core import TENSOR_TYPE, TensorOrder
from ....utils import register_tokenizer
from ...operands import LearnOperand, LearnOperandMixin, OutputType
from .dmatrix import ToDMatrix, check_data

try:
    from xgboost import Booster

    register_tokenizer(Booster, pickle.dumps)
except ImportError:
    pass


class XGBPredict(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.XGBOOST_PREDICT

    _data = KeyField('data')
    _model = BytesField('model', on_serialize=pickle.dumps, on_deserialize=pickle.loads)

    def __init__(self, data=None, model=None, output_types=None, gpu=None, **kw):
        super().__init__(_data=data, _model=model, _gpu=gpu, _output_types=output_types, **kw)

    @property
    def data(self):
        return self._data

    @property
    def model(self):
        return self._model

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._data = self._inputs[0]

    def __call__(self):
        num_class = self._model.attr('num_class')
        if num_class is not None:
            num_class = int(num_class)
        if num_class is not None:
            shape = (len(self._data), num_class)
        else:
            shape = (len(self._data),)
        if self.output_types[0] == OutputType.tensor:
            # tensor
            return self.new_tileable([self._data], shape=shape, dtype=np.dtype(np.float32),
                                     order=TensorOrder.C_ORDER)
        elif self.output_types[0] == OutputType.dataframe:
            # dataframe
            dtypes = pd.DataFrame(np.random.rand(0, num_class), dtype=np.float32).dtypes
            return self.new_tileable([self._data], shape=shape, dtypes=dtypes,
                                     columns_value=parse_index(dtypes.index),
                                     index_value=self._data.index_value)
        else:
            # series
            return self.new_tileable([self._data], shape=shape, index_value=self._data.index_value,
                                     name='predictions', dtype=np.dtype(np.float32))

    @classmethod
    def tile(cls, op):
        out = op.outputs[0]
        out_chunks = []
        data = op.data
        if data.chunk_shape[1] > 1:
            data = data.rechunk({1: op.data.shape[1]})._inplace_tile()
        for in_chunk in data.chunks:
            chunk_op = op.copy().reset_key()
            chunk_index = (in_chunk.index[0],)
            if op.model.attr('num_class'):
                chunk_shape = (len(in_chunk), int(op.model.attr('num_class')))
                chunk_index += (0,)
            else:
                chunk_shape = (len(in_chunk),)
            if op.output_types[0] == OutputType.tensor:
                out_chunk = chunk_op.new_chunk([in_chunk], shape=chunk_shape,
                                               dtype=out.dtype,
                                               order=out.order, index=chunk_index)
            elif op.output_types[0] == OutputType.dataframe:
                # dataframe chunk
                out_chunk = chunk_op.new_chunk([in_chunk], shape=chunk_shape,
                                               dtypes=data.dtypes,
                                               columns_value=data.columns_value,
                                               index_value=in_chunk.index_value,
                                               index=chunk_index)
            else:
                # series chunk
                out_chunk = chunk_op.new_chunk([in_chunk], shape=chunk_shape,
                                               dtype=out.dtype,
                                               index_value=in_chunk.index_value,
                                               name=out.name, index=chunk_index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out.params
        params['chunks'] = out_chunks
        nsplits = (data.nsplits[0],)
        if out.ndim > 1:
            nsplits += ((out.shape[1],),)
        params['nsplits'] = nsplits
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        from xgboost import DMatrix

        raw_data = data = ctx[op.data.key]
        if isinstance(data, tuple):
            data = ToDMatrix.get_xgb_dmatrix(data)
        else:
            data = DMatrix(data)
        result = op.model.predict(data)

        if isinstance(op.outputs[0], DATAFRAME_CHUNK_TYPE):
            result = pd.DataFrame(result, index=raw_data.index)
        elif isinstance(op.outputs[0], SERIES_CHUNK_TYPE):
            result = pd.Series(result, index=raw_data.index, name='predictions')

        ctx[op.outputs[0].key] = result


def predict(model, data, session=None, run_kwargs=None, run=True):
    from xgboost import Booster

    data = check_data(data)
    if not isinstance(model, Booster):
        raise TypeError(f'model has to be a xgboost.Booster, got {type(model)} instead')

    num_class = model.attr('num_class')
    if isinstance(data, TENSOR_TYPE):
        output_types = [OutputType.tensor]
    elif num_class is not None:
        output_types = [OutputType.dataframe]
    else:
        output_types = [OutputType.series]

    op = XGBPredict(data=data, model=model, gpu=data.op.gpu, output_types=output_types)
    result = op()
    if run:
        result.execute(session=session, **(run_kwargs or dict()))
    return result



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from ..actors import FunctionActor
from ..node_info import gather_node_info

logger = logging.getLogger(__name__)


class NodeInfoActor(FunctionActor):
    def __init__(self):
        super().__init__()
        self._node_info = None

    @classmethod
    def default_uid(cls):
        return 's:h1:' + cls.__name__

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        self.ref().gather_info()

    def gather_info(self):
        self._node_info = gather_node_info()
        self.ref().gather_info(_tell=True, _delay=1)

    def get_info(self):
        return self._node_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
import operator
from functools import reduce
from collections.abc import Iterable

from .array import SparseNDArray
from .matrix import SparseMatrix
from .vector import SparseVector
from .core import issparse, get_sparse_module

from .coo import COONDArray


def asarray(x, shape=None):
    from .core import issparse

    if issparse(x):
        return SparseNDArray(x, shape=shape)

    return x


def add(a, b, **_):
    try:
        return a + b
    except TypeError:
        if hasattr(b, '__radd__'):
            return b.__radd__(a)
        raise


def subtract(a, b, **_):
    try:
        return a - b
    except TypeError:
        if hasattr(b, '__rsub__'):
            return b.__rsub__(a)
        raise


def multiply(a, b, **_):
    try:
        return a * b
    except TypeError:
        if hasattr(b, '__rmul__'):
            return b.__rmul__(a)
        raise


def divide(a, b, **_):
    try:
        return a / b
    except TypeError:
        if hasattr(b, '__rdiv__'):
            return b.__rdiv__(a)
        raise


def true_divide(a, b, **_):
    try:
        return a / b
    except TypeError:
        if hasattr(b, '__rtruediv__'):
            return b.__rtruediv__(a)
        raise


def floor_divide(a, b, **_):
    try:
        return a // b
    except TypeError:
        if hasattr(b, '__rfloordiv__'):
            return b.__rfloordiv__(a)
        raise


def power(a, b, **_):
    try:
        return a ** b
    except TypeError:
        if hasattr(b, '__rpow__'):
            return b.__rpow__(a)
        raise


def mod(a, b, **_):
    try:
        return a % b
    except TypeError:
        if hasattr(b, '__rmod__'):
            return b.__rmod__(a)
        raise


def _call_bin(method, a, b, **kwargs):
    from .core import get_array_module, cp, issparse
    from .array import call_sparse_binary_scalar

    # order does not take effect for sparse
    kwargs.pop('order', None)
    if hasattr(a, method):
        res = getattr(a, method)(b, **kwargs)
    elif get_array_module(a).isscalar(a):
        res = call_sparse_binary_scalar(method, a, b, **kwargs)
    else:
        assert get_array_module(a) == get_array_module(b)
        xp = get_array_module(a)
        try:
            res = getattr(xp, method)(a, b, **kwargs)
        except TypeError:
            if xp is cp and issparse(b):
                res = getattr(xp, method)(a, b.toarray(), **kwargs)
            else:
                raise

    if res is NotImplemented:
        raise NotImplementedError

    return res


def _call_unary(method, x, *args, **kwargs):
    from .core import get_array_module

    # order does not take effect for sparse
    kwargs.pop('order', None)
    if hasattr(x, method):
        res = getattr(x, method)(*args, **kwargs)
    else:
        xp = get_array_module(x)
        res = getattr(xp, method)(x, *args, **kwargs)

    if res is NotImplemented:
        raise NotImplementedError

    return res


def float_power(a, b, **kw):
    return _call_bin('float_power', a, b, **kw)


def fmod(a, b, **kw):
    return _call_bin('fmod', a, b, **kw)


def logaddexp(a, b, **kw):
    return _call_bin('logaddexp', a, b, **kw)


def logaddexp2(a, b, **kw):
    return _call_bin('logaddexp2', a, b, **kw)


def negative(x, **_):
    return -x


def positive(x, **_):
    return operator.pos(x)


def absolute(x, **_):
    return builtins.abs(x)


abs = absolute


def fabs(x, **kw):
    return _call_unary('fabs', x, **kw)


def rint(x, **kw):
    return _call_unary('rint', x, **kw)


def sign(x, **kw):
    return _call_unary('sign', x, **kw)


def conj(x, **kw):
    return _call_unary('conj', x, **kw)


def exp(x, **kw):
    return _call_unary('exp', x, **kw)


def exp2(x, **kw):
    return _call_unary('exp2', x, **kw)


def log(x, **kw):
    return _call_unary('log', x, **kw)


def log2(x, **kw):
    return _call_unary('log2', x, **kw)


def log10(x, **kw):
    return _call_unary('log10', x, **kw)


def expm1(x, **kw):
    return _call_unary('expm1', x, **kw)


def log1p(x, **kw):
    return _call_unary('log1p', x, **kw)


def sqrt(x, **kw):
    return _call_unary('sqrt', x, **kw)


def square(x, **kw):
    return _call_unary('square', x, **kw)


def cbrt(x, **kw):
    return _call_unary('cbrt', x, **kw)


def reciprocal(x, **kw):
    return _call_unary('reciprocal', x, **kw)


def gammaln(x, **kw):
    return _call_unary('gammaln', x, **kw)


def erf(x, **kw):
    return _call_unary('erf', x, **kw)


def entr(x, **kw):
    return _call_unary('entr', x, **kw)


def rel_entr(x1, x2, **kw):
    return _call_bin('rel_entr', x1, x2, **kw)


def equal(a, b, **_):
    try:
        return a == b
    except TypeError:
        return b == a


def not_equal(a, b, **_):
    try:
        return a != b
    except TypeError:
        return b != a


def less(a, b, **_):
    try:
        return a < b
    except TypeError:
        return b > a


def less_equal(a, b, **_):
    try:
        return a <= b
    except TypeError:
        return b >= a


def greater(a, b, **_):
    try:
        return a > b
    except TypeError:
        return b < a


def greater_equal(a, b, **_):
    try:
        return a >= b
    except TypeError:
        return b <= a


def logical_and(a, b, **kw):
    return _call_bin('logical_and', a, b, **kw)


def logical_or(a, b, **kw):
    return _call_bin('logical_or', a, b, **kw)


def logical_xor(a, b, **kw):
    return _call_bin('logical_xor', a, b, **kw)


def logical_not(x, **kw):
    return _call_unary('logical_not', x, **kw)


def isclose(a, b, **kw):
    return _call_bin('isclose', a, b, **kw)


def bitwise_and(a, b, **_):
    try:
        return a & b
    except TypeError:
        return b & a


def bitwise_or(a, b, **_):
    try:
        return a | b
    except TypeError:
        return b | a


def bitwise_xor(a, b, **_):
    try:
        return operator.xor(a, b)
    except TypeError:
        return operator.xor(b, a)


def invert(x, **_):
    return ~x


def left_shift(a, b, **_):
    return a << b


def right_shift(a, b, **_):
    return a >> b


def sin(x, **kw):
    return _call_unary('sin', x, **kw)


def cos(x, **kw):
    return _call_unary('cos', x, **kw)


def tan(x, **kw):
    return _call_unary('tan', x, **kw)


def arcsin(x, **kw):
    return _call_unary('arcsin', x, **kw)


def arccos(x, **kw):
    return _call_unary('arccos', x, **kw)


def arctan(x, **kw):
    return _call_unary('arctan', x, **kw)


def arctan2(a, b, **kw):
    return _call_bin('arctan2', a, b, **kw)


def hypot(a, b, **kw):
    return _call_bin('hypot', a, b, **kw)


def sinh(x, **kw):
    return _call_unary('sinh', x, **kw)


def cosh(x, **kw):
    return _call_unary('cosh', x, **kw)


def tanh(x, **kw):
    return _call_unary('tanh', x, **kw)


def arcsinh(x, **kw):
    return _call_unary('arcsinh', x, **kw)


def arccosh(x, **kw):
    return _call_unary('arccosh', x, **kw)


def around(x, **kw):
    return _call_unary('around', x, **kw)


def arctanh(x, **kw):
    return _call_unary('arctanh', x, **kw)


def deg2rad(x, **kw):
    return _call_unary('deg2rad', x, **kw)


def rad2deg(x, **kw):
    return _call_unary('rad2deg', x, **kw)


def angle(x, **kw):
    return _call_unary('angle', x, **kw)


def isinf(x, **kw):
    return _call_unary('isinf', x, **kw)


def isnan(x, **kw):
    return _call_unary('isnan', x, **kw)


def signbit(x, **kw):
    return _call_unary('signbit', x, **kw)


def dot(a, b, sparse=True, **_):
    from .core import issparse

    if not issparse(a):
        ret = a.dot(b)
        if not sparse:
            return ret
        else:
            xps = get_sparse_module(ret)
            return SparseNDArray(xps.csr_matrix(ret), shape=ret.shape)

    return a.dot(b, sparse=sparse)


def tensordot(a, b, axes=2, sparse=True):
    if isinstance(axes, Iterable):
        a_axes, b_axes = axes
    else:
        a_axes = tuple(range(a.ndim - 1, a.ndim - axes - 1, -1))
        b_axes = tuple(range(0, axes))

    if isinstance(a_axes, Iterable):
        a_axes = tuple(a_axes)
    else:
        a_axes = (a_axes,)
    if isinstance(b_axes, Iterable):
        b_axes = tuple(b_axes)
    else:
        b_axes = (b_axes,)

    if a_axes == (a.ndim - 1,) and b_axes == (b.ndim - 2,):
        return dot(a, b, sparse=sparse)

    if a.ndim == b.ndim == 2:
        if a_axes == (a.ndim - 1,) and b_axes == (b.ndim - 1,):
            # inner product of multiple dims
            return dot(a, b.T, sparse=sparse)

    if a.ndim == 1 or b.ndim == 1:
        return dot(a, b, sparse=sparse)

    raise NotImplementedError


def matmul(a, b, sparse=True, **_):
    return dot(a, b, sparse=sparse)


def concatenate(tensors, axis=0):
    return reduce(lambda a, b: _call_bin('concatenate', a, b, axis=axis), tensors)


def transpose(tensor, axes=None):
    return _call_unary('transpose', tensor, axes=axes)


def swapaxes(tensor, axis1, axis2):
    return _call_unary('swapaxes', tensor, axis1, axis2)


def sum(tensor, axis=None, **kw):
    return _call_unary('sum', tensor, axis=axis, **kw)


def prod(tensor, axis=None, **kw):
    return _call_unary('prod', tensor, axis=axis, **kw)


def amax(tensor, axis=None, **kw):
    return _call_unary('amax', tensor, axis=axis, **kw)


max = amax


def amin(tensor, axis=None, **kw):
    return _call_unary('amin', tensor, axis=axis, **kw)


min = amin


def all(tensor, axis=None, **kw):
    return _call_unary('all', tensor, axis=axis, **kw)


def any(tensor, axis=None, **kw):
    return _call_unary('any', tensor, axis=axis, **kw)


def mean(tensor, axis=None, **kw):
    return _call_unary('mean', tensor, axis=axis, **kw)


def nansum(tensor, axis=None, **kw):
    return _call_unary('nansum', tensor, axis=axis, **kw)


def nanprod(tensor, axis=None, **kw):
    return _call_unary('nanprod', tensor, axis=axis, **kw)


def nanmax(tensor, axis=None, **kw):
    return _call_unary('nanmax', tensor, axis=axis, **kw)


def nanmin(tensor, axis=None, **kw):
    return _call_unary('nanmin', tensor, axis=axis, **kw)


def argmax(tensor, axis=None, **kw):
    return _call_unary('argmax', tensor, axis=axis, **kw)


def nanargmax(tensor, axis=None, **kw):
    return _call_unary('nanargmax', tensor, axis=axis, **kw)


def argmin(tensor, axis=None, **kw):
    return _call_unary('argmin', tensor, axis=axis, **kw)


def nanargmin(tensor, axis=None, **kw):
    return _call_unary('nanargmin', tensor, axis=axis, **kw)


def var(tensor, axis=None, **kw):
    return _call_unary('var', tensor, axis=axis, **kw)


def cumsum(tensor, axis=None, **kw):
    return _call_unary('cumsum', tensor, axis=axis, **kw)


def cumprod(tensor, axis=None, **kw):
    return _call_unary('cumprod', tensor, axis=axis, **kw)


def nancumsum(tensor, axis=None, **kw):
    return _call_unary('nancumsum', tensor, axis=axis, **kw)


def nancumprod(tensor, axis=None, **kw):
    return _call_unary('nancumprod', tensor, axis=axis, **kw)


def count_nonzero(tensor, axis=None, **kw):
    return _call_unary('count_nonzero', tensor, axis=axis, **kw)


def maximum(a, b, **kw):
    return _call_bin('maximum', a, b, **kw)


def minimum(a, b, **kw):
    return _call_bin('minimum', a, b, **kw)


def fmax(a, b, **kw):
    return _call_bin('fmax', a, b, **kw)


def fmin(a, b, **kw):
    return _call_bin('fmin', a, b, **kw)


def floor(x, **kw):
    return _call_unary('floor', x, **kw)


def ceil(x, **kw):
    return _call_unary('ceil', x, **kw)


def trunc(x, **kw):
    return _call_unary('trunc', x, **kw)


def degrees(x, **kw):
    return _call_unary('degrees', x, **kw)


def radians(x, **kw):
    return _call_unary('radians', x, **kw)


def clip(a, a_max, a_min, **kw):
    from .core import get_array_module

    if hasattr(a, 'clip'):
        res = getattr(a, 'clip')(a_max, a_min, **kw)
    else:
        xp = get_array_module(a)
        res = getattr(xp, 'clip')(a, a_max, a_min, **kw)

    if res is NotImplemented:
        raise NotImplementedError

    return res


def iscomplex(x, **kw):
    return _call_unary('iscomplex', x, **kw)


def real(x, **_):
    return x.real


def imag(x, **_):
    return x.imag


def fix(x, **kw):
    return _call_unary('fix', x, **kw)


def i0(x, **kw):
    return _call_unary('i0', x, **kw)


def nan_to_num(x, **kw):
    return _call_unary('nan_to_num', x, **kw)


def copysign(a, b, **kw):
    return _call_bin('copysign', a, b, **kw)


def nextafter(a, b, **kw):
    return _call_bin('nextafter', a, b, **kw)


def spacing(x, **kw):
    return _call_unary('spacing', x, **kw)


def ldexp(a, b, **kw):
    return _call_bin('ldexp', a, b, **kw)


def frexp(x, **kw):
    return _call_unary('frexp', x, **kw)


def modf(x, **kw):
    return _call_unary('modf', x, **kw)


def sinc(x, **kw):
    return _call_unary('sinc', x, **kw)


def isfinite(x, **kw):
    return _call_unary('isfinite', x, **kw)


def isreal(x, **kw):
    return _call_unary('isreal', x, **kw)


def where(cond, x, y):
    if any([i.ndim not in (0, 2) for i in (cond, x, y)]):
        raise NotImplementedError

    from .matrix import where as matrix_where
    return matrix_where(cond, x, y)


def digitize(x, bins, right=False):
    return _call_unary('digitize', x, bins, right)


def repeat(a, repeats, axis=None):
    return _call_unary('repeat', a, repeats, axis=axis)


def fill_diagonal(a, val, wrap=False):
    return _call_unary('fill_diagonal', a, val, wrap=wrap)


def unique(a, return_index=False, return_inverse=False,
           return_counts=False, axis=None):
    return _call_unary('unique', a, return_index=return_index,
                       return_inverse=return_inverse,
                       return_counts=return_counts,
                       axis=axis)


def zeros(shape, dtype=float, gpu=False):
    if len(shape) == 2:
        from .matrix import zeros_sparse_matrix

        return zeros_sparse_matrix(shape, dtype=dtype, gpu=gpu)

    raise NotImplementedError


def ones_like(x):
    from .core import get_array_module

    return get_array_module(x).ones(x.shape)


def diag(v, k=0, gpu=False):
    assert v.ndim in {1, 2}

    from .matrix import diag_sparse_matrix

    return diag_sparse_matrix(v, k=k, gpu=gpu)


def eye(N, M=None, k=0, dtype=float, gpu=False):
    from .matrix import eye_sparse_matrix

    return eye_sparse_matrix(N, M=M, k=k, dtype=dtype, gpu=gpu)


def triu(m, k=0, gpu=False):
    if m.ndim == 2:
        from .matrix import triu_sparse_matrix

        return triu_sparse_matrix(m, k=k, gpu=gpu)

    raise NotImplementedError


def tril(m, k=0, gpu=False):
    if m.ndim == 2:
        from .matrix import tril_sparse_matrix

        return tril_sparse_matrix(m, k=k, gpu=gpu)

    raise NotImplementedError


def lu(m):
    from .matrix import lu_sparse_matrix

    return lu_sparse_matrix(m)


def solve_triangular(a, b, lower=False, sparse=True):
    from .matrix import solve_triangular_sparse_matrix

    return solve_triangular_sparse_matrix(a, b, lower=lower, sparse=sparse)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorSin(TensorUnaryOp):
    _op_type_ = OperandDef.SIN
    _func_name = 'sin'


@infer_dtype(np.sin)
def sin(x, out=None, where=None, **kwargs):
    r"""
    Trigonometric sine, element-wise.

    Parameters
    ----------
    x : array_like
        Angle, in radians (:math:`2 \pi` rad equals 360 degrees).
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : array_like
        The sine of each element of x.

    See Also
    --------
    arcsin, sinh, cos

    Notes
    -----
    The sine is one of the fundamental functions of trigonometry (the
    mathematical study of triangles).  Consider a circle of radius 1
    centered on the origin.  A ray comes in from the :math:`+x` axis, makes
    an angle at the origin (measured counter-clockwise from that axis), and
    departs from the origin.  The :math:`y` coordinate of the outgoing
    ray's intersection with the unit circle is the sine of that angle.  It
    ranges from -1 for :math:`x=3\pi / 2` to +1 for :math:`\pi / 2.`  The
    function has zeroes where the angle is a multiple of :math:`\pi`.
    Sines of angles between :math:`\pi` and :math:`2\pi` are negative.
    The numerous properties of the sine and related functions are included
    in any standard trigonometry text.

    Examples
    --------
    Print sine of one angle:

    >>> import mars.tensor as mt

    >>> mt.sin(mt.pi/2.).execute()
    1.0

    Print sines of an array of angles given in degrees:

    >>> mt.sin(mt.array((0., 30., 45., 60., 90.)) * mt.pi / 180. ).execute()
    array([ 0.        ,  0.5       ,  0.70710678,  0.8660254 ,  1.        ])

    Plot the sine function:

    >>> import matplotlib.pylab as plt
    >>> x = mt.linspace(-mt.pi, mt.pi, 201)
    >>> plt.plot(x.execute(), mt.sin(x).execute())
    >>> plt.xlabel('Angle [rad]')
    >>> plt.ylabel('sin(x)')
    >>> plt.axis('tight')
    >>> plt.show()
    """
    op = TensorSin(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
import enum
import itertools
from operator import attrgetter
from typing import List
from weakref import WeakKeyDictionary, WeakSet, ref

import numpy as np

from .serialize import HasKey, HasData, ValueType, ProviderType, Serializable, AttributeAsDict, \
    TupleField, ListField, DictField, KeyField, BoolField, StringField
from .tiles import Tileable, handler
from .utils import tokenize, AttributeDict, on_serialize_shape, \
    on_deserialize_shape, on_serialize_nsplits, enter_mode, is_build_mode


class Base(HasKey):
    __slots__ = ()
    _no_copy_attrs_ = {'_id'}
    _init_update_key_ = True

    def __init__(self, *args, **kwargs):
        for slot, arg in zip(self.__slots__, args):
            object.__setattr__(self, slot, arg)

        for key, val in kwargs.items():
            object.__setattr__(self, key, val)

        if self._init_update_key_ and (not hasattr(self, '_key') or not self._key):
            self._update_key()
        if not hasattr(self, '_id') or not self._id:
            self._id = str(id(self))

    @property
    def _keys_(self):
        cls = type(self)
        member = '__keys_' + cls.__name__
        try:
            return getattr(cls, member)
        except AttributeError:
            slots = sorted(self.__slots__)
            setattr(cls, member, slots)
            return slots

    @property
    def _values_(self):
        return [getattr(self, k, None) for k in self._keys_
                if k not in self._no_copy_attrs_]

    def __mars_tokenize__(self):
        if hasattr(self, '_key'):
            return self._key
        else:
            return (type(self), *self._values_)

    def _obj_set(self, k, v):
        object.__setattr__(self, k, v)

    def _update_key(self):
        self._obj_set('_key', tokenize(type(self).__name__, *self._values_))
        return self

    def reset_key(self):
        self._obj_set('_key', None)
        return self

    def __copy__(self):
        return self.copy()

    def copy(self):
        return self.copy_to(type(self)(_key=self.key))

    def copy_to(self, target):
        for attr in self.__slots__:
            if (attr.startswith('__') and attr.endswith('__')) or attr in self._no_copy_attrs_:
                # we don't copy id to identify that the copied one is new
                continue
            if hasattr(self, attr):
                setattr(target, attr, getattr(self, attr))

        return target

    def copy_from(self, obj):
        obj.copy_to(self)

    @property
    def key(self):
        return self._key

    @property
    def id(self):
        return self._id


class Entity(HasData):
    __slots__ = ()
    _allow_data_type_ = ()

    def __init__(self, data):
        self._check_data(data)
        self._data = data

    def __dir__(self):
        obj_dir = object.__dir__(self)
        if self._data is not None:
            obj_dir = sorted(set(dir(self._data) + obj_dir))
        return obj_dir

    def __str__(self):
        return self._data.__str__()

    def __repr__(self):
        return self._data.__repr__()

    def _check_data(self, data):
        if data is not None and not isinstance(data, self._allow_data_type_):
            raise TypeError(f'Expect {self._allow_data_type_}, got {type(data)}')

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_data):
        self._check_data(new_data)
        self._data = new_data

    def __copy__(self):
        return self.copy()

    def copy(self):
        return self.copy_to(type(self)(None))

    def copy_to(self, target):
        target.data = self._data
        return target

    def copy_from(self, obj):
        self.data = obj.data

    def tiles(self):
        new_entity = self.copy()
        new_entity.data = handler.tiles(self.data)
        return new_entity

    def _inplace_tile(self):
        return handler.inplace_tile(self)

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def __setattr__(self, key, value):
        try:
            object.__setattr__(self, key, value)
        except AttributeError:
            return setattr(self._data, key, value)


class SerializableWithKey(Base, Serializable):
    _key = StringField('key')
    _id = StringField('id')


class AttributeAsDictKey(Base, AttributeAsDict):
    _key = StringField('key')
    _id = StringField('id')


class EntityData(SerializableWithKey):
    __slots__ = '__weakref__', '_siblings'

    # required fields
    _op = KeyField('op')  # store key of operand here
    # optional fields
    _extra_params = DictField('extra_params', key_type=ValueType.string, on_deserialize=AttributeDict)

    def __init__(self, *args, **kwargs):
        extras = AttributeDict((k, kwargs.pop(k)) for k in set(kwargs) - set(self.__slots__))
        kwargs['_extra_params'] = kwargs.pop('_extra_params', extras)
        super().__init__(*args, **kwargs)

    @property
    def op(self):
        return getattr(self, '_op', None)

    @property
    def inputs(self):
        return self.op.inputs or []

    @inputs.setter
    def inputs(self, new_inputs):
        self.op.inputs = new_inputs

    def is_sparse(self):
        return self.op.is_sparse()

    issparse = is_sparse

    @property
    def extra_params(self):
        return self._extra_params


ENTITY_TYPE = (EntityData, Entity)


class ChunkData(EntityData):
    __slots__ = ()

    # optional fields
    _index = TupleField('index', ValueType.uint32)
    _cached = BoolField('cached')

    def __repr__(self):
        if self.op.stage is None:
            return f'Chunk <op={type(self.op).__name__}, key={self.key}>'
        else:
            return f'Chunk <op={type(self.op).__name__}, stage={self.op.stage.name}, key={self.key}>'

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from .serialize.protos.chunk_pb2 import ChunkDef
            return ChunkDef
        return super().cls(provider)

    @property
    def index(self):
        return getattr(self, '_index', None)

    @property
    def cached(self):
        return getattr(self, '_cached', None)

    @property
    def device(self):
        return self.op.device

    def _update_key(self):
        object.__setattr__(self, '_key', tokenize(
            type(self).__name__, *(getattr(self, k, None) for k in self._keys_ if k != '_index')))


class Chunk(Entity):
    __slots__ = ()
    _allow_data_type_ = (ChunkData,)


CHUNK_TYPE = (ChunkData, Chunk)


class ObjectChunkData(ChunkData):
    # chunk whose data could be any serializable
    __slots__ = ()

    def __init__(self, op=None, index=None, **kw):
        super().__init__(_op=op, _index=index, **kw)

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from .serialize.protos.object_pb2 import ObjectChunkDef
            return ObjectChunkDef
        return super().cls(provider)

    @property
    def params(self):
        # params return the properties which useful to rebuild a new chunk
        return {
            'index': self.index,
        }


class ObjectChunk(Chunk):
    __slots__ = ()
    _allow_data_type_ = (ObjectChunkData,)


class FuseChunkData(ChunkData):
    __slots__ = '_inited',

    _chunk = KeyField('chunk',
                      on_serialize=lambda x: x.data if hasattr(x, 'data') else x)

    def __init__(self, *args, **kwargs):
        self._inited = False
        super().__init__(*args, **kwargs)
        self._extra_params = {}
        self._inited = True

    @property
    def chunk(self):
        return self._chunk

    @property
    def composed(self):
        # for compatibility, just return the topological ordering,
        # once we apply optimization on the subgraph,
        # `composed` is not needed any more and should be removed then.
        assert getattr(self._op, 'fuse_graph', None) is not None
        fuse_graph = self._op.fuse_graph
        return list(fuse_graph.topological_iter())

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from .serialize.protos.fusechunk_pb2 import FuseChunkDef
            return FuseChunkDef
        return super().cls(provider)

    def __getattr__(self, attr):
        if not self._inited:
            return object.__getattribute__(self, attr)
        if attr in self._extra_params:
            return self._extra_params[attr]
        try:
            return getattr(self._chunk, attr)
        except AttributeError:
            return object.__getattribute__(self, attr)

    @property
    def nbytes(self):
        return np.prod(self.shape) * self.dtype.itemsize


class FuseChunk(Chunk):
    __slots__ = ()
    _allow_data_type_ = (FuseChunkData,)


FUSE_CHUNK_TYPE = (FuseChunkData, FuseChunk)


class _ExecutableMixin:
    __slots__ = ()

    def execute(self, session=None, **kw):
        from .session import Session

        if 'fetch' in kw and kw['fetch']:
            raise ValueError('Does not support fetch=True for `.execute()`,'
                             'please use `.fetch()` instead')
        else:
            kw['fetch'] = False

        if session is None:
            session = Session.default_or_local()

        # no more fetch, thus just fire run
        session.run(self, **kw)
        # return Tileable or ExecutableTuple itself
        return self

    def fetch(self, session=None, **kw):
        from .session import Session

        if session is None:
            session = Session.default_or_local()
        return session.fetch(self, **kw)

    def _attach_session(self, session):
        _cleaner.register(self, session)
        self._executed_sessions.append(session)


class _ExecuteAndFetchMixin:
    __slots__ = ()

    def _execute_and_fetch(self, session=None, **kw):
        if session is None and len(self._executed_sessions) > 0:
            session = self._executed_sessions[-1]
        try:
            # fetch first, to reduce the potential cost of submitting a graph
            return self.fetch(session=session)
        except ValueError:
            # not execute before
            return self.execute(session=session, **kw).fetch(session=session)


class _ToObjectMixin(_ExecuteAndFetchMixin):
    __slots__ = ()

    def to_object(self, session=None, **kw):
        return self._execute_and_fetch(session=session, **kw)


class TileableData(EntityData, Tileable, _ExecutableMixin):
    __slots__ = '_cix', '_entities', '_executed_sessions'
    _no_copy_attrs_ = SerializableWithKey._no_copy_attrs_ | {'_cix'}

    # optional fields
    # `nsplits` means the sizes of chunks for each dimension
    _nsplits = TupleField('nsplits', ValueType.tuple(ValueType.uint64),
                          on_serialize=on_serialize_nsplits)

    def __init__(self, *args, **kwargs):
        if kwargs.get('_nsplits', None) is not None:
            kwargs['_nsplits'] = tuple(tuple(s) for s in kwargs['_nsplits'])

        super().__init__(*args, **kwargs)

        if hasattr(self, '_chunks') and self._chunks:
            self._chunks = sorted(self._chunks, key=attrgetter('index'))

        self._entities = WeakSet()
        self._executed_sessions = []

    @property
    def chunk_shape(self):
        if hasattr(self, '_nsplits') and self._nsplits is not None:
            return tuple(map(len, self._nsplits))

    @property
    def chunks(self) -> List["Chunk"]:
        return getattr(self, '_chunks', None)

    @property
    def nsplits(self):
        return getattr(self, '_nsplits', None)

    @nsplits.setter
    def nsplits(self, new_nsplits):
        self._nsplits = new_nsplits

    @property
    def params(self) -> dict:
        # params return the properties which useful to rebuild a new tileable object
        return dict()

    @property
    def cix(self):
        if self.ndim == 0:
            return ChunksIndexer(self)

        try:
            if getattr(self, '_cix', None) is None:
                self._cix = ChunksIndexer(self)
            return self._cix
        except (TypeError, ValueError):
            return ChunksIndexer(self)

    @property
    def entities(self):
        return self._entities

    def is_coarse(self):
        return not hasattr(self, '_chunks') or self._chunks is None or len(self._chunks) == 0

    @enter_mode(build=True)
    def attach(self, entity):
        self._entities.add(entity)

    @enter_mode(build=True)
    def detach(self, entity):
        self._entities.discard(entity)


class TileableEntity(Entity):
    __slots__ = '__weakref__',

    def __init__(self, data):
        super().__init__(data)
        if self._data is not None:
            self._data.attach(self)
            if self._data.op.create_view:
                entity_view_handler.add_observer(self._data.inputs[0], self)

    def __copy__(self):
        return self._view()

    def _view(self):
        return super().copy()

    def copy(self):
        new_op = self.op.copy().reset_key()
        if new_op.create_view:
            # if the operand is a view, make it a copy
            new_op._create_view = False
        new_outs = new_op.new_tileables(self.op.inputs, kws=[t.params for t in self.op.outputs],
                                        output_limit=len(self.op.outputs),
                                        **self._data.extra_params)
        pos = -1
        for i, out in enumerate(self.op.outputs):
            if self._data is out:
                pos = i
                break
        assert pos >= 0
        return new_outs[pos]

    @Entity.data.setter
    def data(self, new_data):
        self._check_data(new_data)
        if self._data is None:
            self._data = new_data
            self._data.attach(self)
        else:
            entity_view_handler.data_changed(self._data, new_data)


TILEABLE_TYPE = (TileableEntity, TileableData)


class HasShapeTileableData(TileableData):
    # required fields
    _shape = TupleField('shape', ValueType.int64,
                        on_serialize=on_serialize_shape, on_deserialize=on_deserialize_shape)

    @property
    def ndim(self):
        return len(self.shape)

    def __len__(self):
        try:
            return self.shape[0]
        except IndexError:
            if is_build_mode():
                return 0
            raise TypeError('len() of unsized object')

    @property
    def shape(self):
        if hasattr(self, '_shape') and self._shape is not None:
            return self._shape
        if hasattr(self, '_nsplits') and self._nsplits is not None:
            self._shape = tuple(builtins.sum(nsplit) for nsplit in self._nsplits)
            return self._shape

    def _update_shape(self, new_shape):
        self._shape = new_shape

    @property
    def size(self):
        return np.prod(self.shape).item()

    @property
    def params(self):
        # params return the properties which useful to rebuild a new tileable object
        return {
            'shape': self.shape
        }

    def _equals(self, o):
        return self is o


class HasShapeTileableEnity(TileableEntity):
    __slots__ = ()

    @property
    def shape(self):
        return self._data.shape

    @property
    def ndim(self):
        return self._data.ndim

    @property
    def size(self):
        return self._data.size

    def execute(self, session=None, **kw):
        self._data.execute(session, **kw)
        return self


class ObjectData(TileableData, _ToObjectMixin):
    __slots__ = ()

    # optional fields
    _chunks = ListField('chunks', ValueType.reference(ObjectChunkData),
                        on_serialize=lambda x: [it.data for it in x] if x is not None else x,
                        on_deserialize=lambda x: [ObjectChunk(it) for it in x] if x is not None else x)

    def __init__(self, op=None, nsplits=None, chunks=None, **kw):
        super().__init__(_op=op, _nsplits=nsplits, _chunks=chunks, **kw)

    def __repr__(self):
        return f'Object <op={type(self.op).__name__}, key={self.key}>'

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from .serialize.protos.object_pb2 import ObjectDef
            return ObjectDef
        return super().cls(provider)

    @property
    def params(self):
        # params return the properties which useful to rebuild a new tileable object
        return {
        }


class Object(Entity, _ToObjectMixin):
    __slots__ = ()
    _allow_data_type_ = (ObjectData,)


OBJECT_TYPE = (Object, ObjectData)
OBJECT_CHUNK_TYPE = (ObjectChunk, ObjectChunkData)


class ChunksIndexer(object):
    __slots__ = '_tileable',

    def __init__(self, tileable):
        self._tileable = tileable

    def __getitem__(self, item):
        """
        The indices for `cix` can be [x, y] or [x, :]. For the former the result will be
        a single chunk, and for the later the result will be a list of chunks (flattened).

        The length of indices must be the same with `chunk_shape` of tileable.
        """
        if isinstance(item, tuple):
            if len(item) == 0 and self._tileable.is_scalar():
                return self._tileable.chunks[0]
            if len(item) != self._tileable.ndim:
                raise ValueError(f'Cannot get chunk by {item}, expect length {self._tileable.ndim}')
            slices, singleton = [], True
            for it, dim in zip(item, self._tileable.chunk_shape):
                if isinstance(it, slice):
                    slices.append(range(dim)[it])
                    singleton = False
                elif np.issubdtype(type(it), np.integer):
                    slices.append([it if it >= 0 else dim + it])
                else:
                    raise TypeError(f'Cannot get chunk by {it}, invalid value has type {type(it)}')

            indexes = tuple(zip(*itertools.product(*slices)))

            flat_index = np.ravel_multi_index(indexes, self._tileable.chunk_shape)
            if singleton:
                return self._tileable._chunks[flat_index[0]]
            else:
                return [self._tileable._chunks[idx] for idx in flat_index]

        raise ValueError(f'Cannot get {type(self._tileable).__name__} chunk by {item}')


class ExecutableTuple(tuple, _ExecutableMixin, _ToObjectMixin):
    def __init__(self, *_):
        super().__init__()
        self._executed_sessions = []

    def execute(self, session=None, **kw):
        if len(self) == 0:
            return self
        return super().execute(session=session, **kw)

    def fetch(self, session=None, **kw):
        if len(self) == 0:
            return tuple()
        return super().fetch(session=session, **kw)

    def _attach_session(self, session):
        super()._attach_session(session)
        for t in self:
            if hasattr(t, '_attach_session'):
                t._attach_session(session)


class _TileableSession:
    def __init__(self, tensor, session):
        key = tensor.key, tensor.id

        def cb(_, sess=ref(session)):
            s = sess()
            if s:
                s.decref(key)
        self._tensor = ref(tensor, cb)


class _TileableDataCleaner:
    def __init__(self):
        self._tileable_to_sessions = WeakKeyDictionary()

    @enter_mode(build=True)
    def register(self, tensor, session):
        if tensor in self._tileable_to_sessions:
            self._tileable_to_sessions[tensor].append(_TileableSession(tensor, session))
        else:
            self._tileable_to_sessions[tensor] = [_TileableSession(tensor, session)]


# we don't use __del__ to avoid potential Circular reference
_cleaner = _TileableDataCleaner()


class EntityDataModificationHandler:
    def __init__(self):
        self._data_to_entities = WeakKeyDictionary()

    def _add_observer(self, data, entity):
        # only tileable data should be considered
        assert isinstance(data, TileableData)
        assert isinstance(entity, TileableEntity)

        if data not in self._data_to_entities:
            self._data_to_entities[data] = WeakSet()

        self._data_to_entities[data].add(entity)

    @enter_mode(build=True)
    def add_observer(self, data, entity):
        self._add_observer(data, entity)

    def _update_observe_data(self, observer, data, new_data):
        self._data_to_entities.get(data, set()).discard(observer)
        self._add_observer(new_data, observer)

    @staticmethod
    def _set_data(entity, data):
        entity._data.detach(entity)
        entity._data = data
        data.attach(entity)

    @staticmethod
    def _get_data(obj):
        return obj.data if isinstance(obj, Entity) else obj

    @enter_mode(build=True)
    def data_changed(self, old_data, new_data):
        notified = set()
        processed_data = set()
        old_to_new = {old_data: new_data}
        q = [old_data]
        while len(q) > 0:
            data = q.pop()

            # handle entities
            for entity in data.entities:
                self._set_data(entity, old_to_new[data])
                notified.add(entity)

            observers = {ob for ob in self._data_to_entities.pop(data, set())
                         if ob not in notified}
            for ob in observers:
                new_data = self._get_data(ob.op.on_input_modify(old_to_new[data]))
                old_data = ob.data
                self._update_observe_data(ob, ob.data, new_data)
                old_to_new[old_data] = new_data
                if old_data not in processed_data:
                    q.append(old_data)
                    processed_data.add(old_data)
                notified.add(ob)

            if data.op.create_view:
                old_input_data = data.inputs[0]
                new_input_data = self._get_data(data.op.on_output_modify(old_to_new[data]))
                old_to_new[old_input_data] = new_input_data
                if old_input_data not in processed_data:
                    q.append(old_input_data)
                    processed_data.add(old_input_data)


entity_view_handler = EntityDataModificationHandler()


class OutputType(enum.Enum):
    object = 1
    tensor = 2
    dataframe = 3
    series = 4
    index = 5
    scalar = 6
    categorical = 7
    dataframe_groupby = 8
    series_groupby = 9

    @classmethod
    def serialize_list(cls, output_types):
        return [ot.value for ot in output_types] if output_types is not None else None

    @classmethod
    def deserialize_list(cls, output_types):
        return [cls(ot) for ot in output_types] if output_types is not None else None


_OUTPUT_TYPE_TO_CHUNK_TYPES = {OutputType.object: OBJECT_CHUNK_TYPE}
_OUTPUT_TYPE_TO_TILEABLE_TYPES = {OutputType.object: OBJECT_TYPE}


def register_output_types(output_type, tileable_types, chunk_types):
    _OUTPUT_TYPE_TO_TILEABLE_TYPES[output_type] = tileable_types
    _OUTPUT_TYPE_TO_CHUNK_TYPES[output_type] = chunk_types


def get_tileable_types(output_type):
    return _OUTPUT_TYPE_TO_TILEABLE_TYPES[output_type]


def get_chunk_types(output_type):
    return _OUTPUT_TYPE_TO_CHUNK_TYPES[output_type]


def get_output_types(*objs, unknown_as=None):
    output_types = []
    for obj in objs:
        if obj is None:
            continue
        for tp in OutputType.__members__.values():
            try:
                tileable_types = _OUTPUT_TYPE_TO_TILEABLE_TYPES[tp]
                chunk_types = _OUTPUT_TYPE_TO_CHUNK_TYPES[tp]
                if isinstance(obj, (tileable_types, chunk_types)):
                    output_types.append(tp)
                    break
            except KeyError:
                continue
        else:
            if unknown_as is not None:
                output_types.append(unknown_as)
            else:  # pragma: no cover
                raise TypeError('Output can only be tensor, dataframe or series')
    return output_types



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
import sys
import pickle
import random
import struct
import itertools
from collections import OrderedDict

import gevent
import gevent.queue
import gevent.lock
import gevent.event
import gevent.pool
import gevent.socket
import gevent.server
import gevent.subprocess
import gevent.fileobject
import gevent.threadpool
from gevent._tblib import _init as gevent_init_tblib

from ...lib import gipc
from ..errors import ActorPoolNotStarted, ActorNotExist, ActorAlreadyExist
from ..distributor cimport Distributor
from ..core cimport ActorRef, Actor
from ..cluster cimport ClusterInfo
from .messages cimport pack_send_message, pack_tell_message, pack_create_actor_message, \
    pack_destroy_actor_message, pack_result_message, unpack_send_message, unpack_tell_message, \
    unpack_create_actor_message, unpack_destroy_actor_message, unpack_result_message, \
    pack_has_actor_message, unpack_has_actor_message, pack_error_message, unpack_error_message, \
    unpack_message_type_value, unpack_message_type, unpack_message_id, read_remote_message, \
    get_index, MessageType
from .messages import write_remote_message
from .utils cimport new_actor_id
from .utils import create_actor_ref

gevent_init_tblib()

cdef int REMOTE_FROM_INDEX = -2
cdef int UNKNOWN_TO_INDEX = -1
cpdef int REMOTE_DEFAULT_PARALLEL = 50  # parallel connection at most
cpdef int REMOTE_MAX_CONNECTION = 200  # most connections

_inaction_encoder = _inaction_decoder = None


cdef class MessageContext:
    cdef public object message
    cdef public object async_result

    def __init__(self, message):
        self.message = message
        self.async_result = gevent.event.AsyncResult()

    cpdef result(self):
        return self.async_result.result()


cdef class ActorExecutionContext:
    cdef public object actor
    cdef public object lock
    cdef public object inbox
    cdef public object comm

    def __init__(self, actor, comm):
        self.actor = weakref.ref(actor)
        self.lock = gevent.lock.RLock()
        self.inbox = gevent.queue.Queue()
        self.comm = comm

    cpdef object fire_run(self):
        cdef MessageContext message_ctx

        with self.lock:
            message_ctx = self.inbox.get()
            actor = self.actor()

            try:
                res = actor.on_receive(message_ctx.message)
                message_ctx.async_result.set_result(res)
                return res
            except:
                t, ex, tb = sys.exc_info()
                message_ctx.async_result.set_exception(ex, exc_info=(t, ex, tb))
                raise

    cpdef void receive(self, MessageContext message_ctx):
        self.inbox.put(message_ctx)


class GeventThreadPool:
    def __init__(self, num_threads):
        self._pool = gevent.threadpool.ThreadPool(num_threads)

    @staticmethod
    def _wrap_watch(fn):
        # Each time a function is submitted, a gevent greenlet may be created,
        # this is common especially for Mars actor,
        # but there would be no other greenlet to switch to,
        # LoopExit will be raised, in order to prevent from this,
        # we create a greenlet to watch the result

        def check(event):
            delay = 0.0005
            while not event.is_set():
                event.wait(delay)
                delay = min(delay * 2, .05)

        def inner(*args, **kwargs):
            event = gevent.event.Event()
            greenlet = gevent.spawn(check, event)
            result = fn(*args, **kwargs)
            event.set()
            greenlet.join()
            return result

        return inner

    def submit(self, fn, *args, **kwargs):
        wrapped_fn = self._wrap_watch(fn)
        return self._pool.spawn(wrapped_fn, *args, **kwargs)


cdef class ActorContext:
    """
    Hold by an actor.

    actor = Actor()
    actor.ctx = ActorContext(comm)

    provide methods:

    create_actor
    actor_ref
    """
    cdef object _comm

    def __init__(self, comm):
        self._comm = comm

    @property
    def index(self):
        return self._comm.index

    @property
    def distributor(self):
        return self._comm.distributor

    def create_actor(self, actor_cls, *args, **kwargs):
        cdef bint wait
        cdef object address
        cdef object uid
        cdef object ref

        wait = kwargs.pop('wait', True)
        address = kwargs.pop('address', None)
        uid = kwargs.pop('uid', None)

        if wait:
            ref = self._comm.create_actor(address, uid, actor_cls, *args, **kwargs)
            ref.ctx = self
            return ref

        def callback(ret):
            ret.ctx = self

        kwargs['wait'] = False
        kwargs['callback'] = callback
        return self._comm.create_actor(address, uid, actor_cls, *args, **kwargs)

    def destroy_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        return self._comm.destroy_actor(actor_ref, wait=wait, callback=callback)

    def has_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        return self._comm.has_actor(actor_ref, wait=wait, callback=callback)

    def actor_ref(self, *args, **kwargs):
        cdef ActorRef ref

        ref = create_actor_ref(*args, **kwargs)
        ref.ctx = self
        return ref

    def send(self, ActorRef actor_ref, object message, bint wait=True, object callback=None):
        return self._comm.send(actor_ref, message, wait=wait, callback=callback)

    def tell(self, ActorRef actor_ref, object message, object delay=None,
             bint wait=True, object callback=None):
        return self._comm.tell(actor_ref, message, delay=delay,
                               wait=wait, callback=callback)

    @staticmethod
    def event():
        return gevent.event.Event()

    @staticmethod
    def sleep(seconds):
        gevent.sleep(seconds)

    @staticmethod
    def fileobject(fobj, mode='rb', bufsize=-1, close=True):
        return gevent.fileobject.FileObject(fobj, mode=mode, bufsize=bufsize, close=close)

    @staticmethod
    def popen(*args, **kwargs):
        return gevent.subprocess.Popen(*args, **kwargs)

    @staticmethod
    def threadpool(size):
        return GeventThreadPool(size)

    @staticmethod
    def asyncpool(size=None):
        return gevent.pool.Pool(size)

    @staticmethod
    def lock():
        return gevent.lock.Semaphore()


cdef class LocalActorPool:
    """
    manage local actors, this pool is not aware of the multi-processes or cluster
    """

    cdef public dict actors
    cdef public object comm_proxy
    cdef public str address
    cdef public int index

    def __init__(self, str address=None, index=0):
        self.actors = dict()
        self.comm_proxy = None
        self.address = address
        self.index = index

    @property
    def comm(self):
        return self.comm_proxy()

    cpdef set_comm(self, object comm):
        self.comm_proxy = weakref.ref(comm)

    cpdef ActorExecutionContext get_actor_execution_ctx(self, object actor_uid):
        try:
            return self.actors[actor_uid][1]
        except KeyError:
            raise ActorNotExist(f'Actor {actor_uid} does not exist')

    def create_actor(self, object actor_cls, object uid, *args, **kw):
        cdef Actor actor

        actor = actor_cls(*args, **kw)
        actor.address = self.address
        actor.uid = uid
        actor_execution_ctx = ActorExecutionContext(actor, self.comm)
        actor.ctx = ActorContext(actor_execution_ctx.comm)
        if uid in self.actors:
            raise ActorAlreadyExist(f'Actor {actor.uid} already exist, cannot create')
        self.actors[uid] = (actor, actor_execution_ctx)
        actor.post_create()
        return ActorRef(self.address, uid)

    cpdef bint has_actor(self, object actor_uid) except -128:
        if actor_uid not in self.actors:
            return False
        return True

    cpdef destroy_actor(self, object actor_uid):
        cdef Actor actor

        try:
            actor, _ = self.actors[actor_uid]
        except KeyError:
            raise ActorNotExist(f'Actor {actor_uid} does not exist')
        actor.pre_destroy()
        del self.actors[actor_uid]
        return actor_uid


cdef class AsyncHandler:
    cdef dict async_results

    def __init__(self):
        self.async_results = dict()

    cpdef object submit(self, bytes unique_id):
        cdef object ar

        ar = self.async_results[unique_id] = gevent.event.AsyncResult()

        # wait for result
        ar.wait()
        del self.async_results[unique_id]

        return ar.result()

    cpdef object future(self, bytes unique_id, object future, object callback=None):
        cdef object ar

        ar = self.async_results[unique_id] = gevent.event.AsyncResult()

        # wait for result
        ar.wait()
        del self.async_results[unique_id]

        if callback and ar.successful():
            callback(ar.result())

        future(ar)

    cpdef object wait(self, bytes unique_id):
        cdef object ar

        ar = self.async_results[unique_id] = gevent.event.AsyncResult()

        # wait for result
        ar.wait()
        del self.async_results[unique_id]

        if ar.successful():
            return ar.result()
        else:
            return ar.exception()

    cpdef void got(self, bytes unique_id, object result):
        # receive event handling
        self.async_results[unique_id].set_result(result)

    cpdef void err(self, bytes unique_id, object t, object ex, object tb):
        self.async_results[unique_id].set_exception(ex, exc_info=(t, ex, tb))


cdef class Connection:
    cdef public object conn
    cdef public object lock

    def __init__(self, conn, lock):
        self.conn = conn
        self.lock = lock

    def __enter__(self):
        return self.conn

    def __exit__(self, *_):
        self.lock.release()

    cpdef void release(self):
        self.lock.release()


class Connections(object):
    global_lock = gevent.lock.Semaphore()
    addrs = 0

    def __init__(self, address):
        if isinstance(address, str):
            self.address = address.split(':', 1)
        else:
            self.address = address

        with self.global_lock:
            Connections.addrs += 1

        self.lock = gevent.lock.Semaphore()
        self.conn_locks = OrderedDict()

    @property
    def conn(self):
        return [conn_lock[0] for conn_lock in self.conn_locks.values()]

    def _connect(self, conn, lock):
        return Connection(conn, lock)

    def got_broken_pipe(self, fd):
        del self.conn_locks[fd]

    def connect(self):
        cdef int maxlen
        cdef object conn
        cdef object lock

        with self.lock:
            for conn, lock in self.conn_locks.values():
                # try to reuse the connections before
                locked = lock.acquire(blocking=False)
                if not locked:
                    continue
                return self._connect(conn, lock)

            maxlen = max(REMOTE_MAX_CONNECTION // Connections.addrs, 1)

            if len(self.conn) < maxlen:
                # create a new connection
                lock = gevent.lock.Semaphore()
                lock.acquire()
                conn = gevent.socket.create_connection(self.address)
                self.conn_locks[conn.fileno()] = (conn, lock)
                return self._connect(conn, lock)

            def close(c, lk):
                with lk:
                    c.close()

            ps = [gevent.spawn(close, c, l) for c, l in
                  itertools.islice(self.conn_locks.values(), maxlen, len(self.conn_locks))]

            i = random.randint(0, maxlen - 1)
            fd = next(itertools.islice(self.conn_locks.keys(), i, i + 1))
            conn, lock = self.conn_locks[fd]
            lock.acquire()

            # wait for conn finished
            gevent.joinall(ps)
            self.conn_locks = OrderedDict(itertools.islice(self.conn_locks.items(), maxlen))

            return self._connect(conn, lock)

    def __del__(self):
        for c, _ in self.conn_locks.values():
            try:
                c.close()
            except:  # pragma: no cover
                pass


cdef class ActorRemoteHelper:
    """
    Used to handle remote operations, like deliver create_actor, destroy_actor, send etc to remote,
    and handle the response.

    An ActorRemoteHelper instance can be hold by Communicator and ActorClient.
    """
    index = REMOTE_FROM_INDEX  # means remote

    cdef object _parallel
    cdef object _pool
    cdef dict _connections
    cdef object _lock

    def __init__(self, parallel=None):
        self._parallel = parallel if parallel is not None else REMOTE_DEFAULT_PARALLEL
        self._pool = gevent.pool.Pool(self._parallel)
        self._connections = dict()
        self._lock = gevent.lock.RLock()

    cdef object _new_connection(self, str address):
        with self._lock:
            if address not in self._connections:
                connections = Connections(address)
                self._connections[address] = connections

            return self._connections[address].connect()

    cpdef object _send_remote(self, str address, object binary):
        cdef bytes res_binary
        cdef object message_type

        with self._new_connection(address) as sock:
            try:
                write_remote_message(sock.sendall, *binary)
                res_binary = read_remote_message(sock.recv)
                message_type = unpack_message_type(res_binary)
                if message_type == MessageType.error:
                    error_message = unpack_error_message(res_binary)
                    raise error_message.error.with_traceback(error_message.traceback) from None
                else:
                    assert message_type == MessageType.result
                    return unpack_result_message(res_binary).result
            except BrokenPipeError:
                self._connections[address].got_broken_pipe(sock.fileno())
                raise

    def create_actor(self, str address, object uid, object actor_cls, *args, **kwargs):
        cdef bint wait
        cdef object callback
        cdef tuple binaries
        cdef ActorRef actor_ref
        cdef object future

        wait = kwargs.pop('wait', True)
        callback = kwargs.pop('callback', None)

        try:
            binaries = pack_create_actor_message(
                ActorRemoteHelper.index, UNKNOWN_TO_INDEX,
                ActorRef(address, uid), actor_cls, args, kwargs)[1:]
        except (AttributeError, pickle.PickleError):
            raise pickle.PicklingError(f'Unable to pickle {actor_cls}(*{args!r}, **{kwargs!r})')

        if wait:
            actor_ref = self._pool.apply(self._send_remote, (address, binaries))
            actor_ref.ctx = ActorContext(self)
            return actor_ref

        # return future
        future = gevent.event.AsyncResult()

        def on_success(g):
            actor_ref = g.value
            actor_ref.ctx = ActorContext(self)
            if callback is not None:
                callback(actor_ref)
            future.set_result(actor_ref)

        def on_failure(g):
            try:
                g.get()
            except:
                t, ex, tb = sys.exc_info()
                future.set_exception(ex, exc_info=(t, ex, tb))

        p = self._pool.apply_async(self._send_remote, (address, binaries))
        p.link_value(on_success)
        p.link_exception(on_failure)
        return future

    def _async_run(self, str address, object binaries, object callback=None):
        cdef object future

        # return future
        future = gevent.event.AsyncResult()

        def on_success(g):
            ret = g.value
            if callback is not None:
                callback(ret)
            future.set_result(ret)

        def on_failure(g):
            try:
                g.get()
            except:
                t, ex, tb = sys.exc_info()
                future.set_exception(ex, exc_info=(t, ex, tb))

        p = self._pool.apply_async(self._send_remote, (address, binaries))
        p.link_value(on_success)
        p.link_exception(on_failure)
        return future

    cpdef object destroy_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        cdef tuple binaries

        binaries = pack_destroy_actor_message(ActorRemoteHelper.index, UNKNOWN_TO_INDEX, actor_ref)[1:]

        if wait:
            return self._pool.apply(self._send_remote, (actor_ref.address, binaries))

        # return future
        return self._async_run(actor_ref.address, binaries, callback=callback)

    cpdef object has_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        cdef tuple binaries

        binaries = pack_has_actor_message(ActorRemoteHelper.index, UNKNOWN_TO_INDEX, actor_ref)[1:]

        if wait:
            return self._pool.apply(self._send_remote, (actor_ref.address, binaries))

        # return future
        return self._async_run(actor_ref.address, binaries, callback=callback)

    def actor_ref(self, *args, **kwargs):
        cdef ActorRef ref

        ref = create_actor_ref(*args, **kwargs)
        ref.ctx = self
        return ref

    cpdef _send(self, ActorRef actor_ref, object message, bint wait_response=True,
                      bint wait=True, object callback=None):
        cdef object func
        cdef list binaries

        try:
            if wait_response:
                binaries = pack_send_message(ActorRemoteHelper.index, UNKNOWN_TO_INDEX, actor_ref, message)[1:]
            else:
                binaries = pack_tell_message(ActorRemoteHelper.index, UNKNOWN_TO_INDEX, actor_ref, message)[1:]
        except (AttributeError, pickle.PickleError):
            raise pickle.PicklingError(f'Unable to pickle message: {message!r}')

        if wait:
            return self._pool.apply(self._send_remote, (actor_ref.address, binaries))

        # return future
        return self._async_run(actor_ref.address, binaries, callback=callback)

    cpdef send(self, ActorRef actor_ref, object message, bint wait=True, object callback=None):
        return self._send(actor_ref, message, wait_response=True, wait=wait, callback=callback)

    cpdef tell(self, ActorRef actor_ref, object message, object delay=None,
                      bint wait=True, object callback=None):
        if delay is not None:
            return gevent.spawn_later(delay, self._send, actor_ref, message, wait_response=False,
                                      wait=wait, callback=callback)

        return self._send(actor_ref, message, wait_response=False, wait=wait, callback=callback)


cdef class Communicator(AsyncHandler):
    """
    Communicator, decide to send message to
    1) local actor pool, just spawn a greenlet to call the ActorExecutionContext.fire_run
    2) local dispatcher, mainly send message to another process, no pipe means no local dispatcher
    3) remote dispatcher which serves a stream handler

    Also, communicator receives message from dispatcher which received from different process or remote.
    """

    cdef public int index
    cdef public Distributor distributor

    cdef object pool
    cdef ClusterInfo cluster_info
    cdef object pipe
    cdef object remote_handler
    cdef object running
    cdef dict _handlers
    cdef object __weakref__

    def __init__(self, pool, ClusterInfo cluster_info, pipe, Distributor distributor=None, parallel=None):
        super().__init__()
        AsyncHandler.__init__(self)

        self.pool = pool
        self.index = self.pool.index
        self.cluster_info = cluster_info
        self.pipe = pipe
        if distributor is None:
            self.distributor = Distributor(self.cluster_info.n_process)
        else:
            self.distributor = distributor
        self.remote_handler = ActorRemoteHelper(parallel=parallel)

        self.running = gevent.event.Event()
        self.running.set()

        self._handlers = {
            MessageType.send_all: self._on_receive_send,
            MessageType.tell_all: self._on_receive_tell,
            MessageType.create_actor: self._on_receive_create_actor,
            MessageType.destroy_actor: self._on_receive_destroy_actor,
            MessageType.has_actor: self._on_receive_has_actor,
            MessageType.result: self._on_receive_result,
            MessageType.error: self._on_receive_error,
        }

    cpdef _dispatch(self, local_func, redirect_func, remote_func, ActorRef actor_ref, args, kwargs):
        cdef int send_to_index
        if actor_ref.address is None or actor_ref.address == self.cluster_info.address \
                or actor_ref.address == self.cluster_info.advertise_address:
            if self.cluster_info.n_process == 1:
                send_to_index = self.index
            else:
                send_to_index = self.distributor.distribute(actor_ref.uid)
            if send_to_index == self.index:
                # send to local actor_pool
                return local_func(*args, **kwargs)
            else:
                return redirect_func(send_to_index, *args, **kwargs)
        else:
            if self.cluster_info.standalone:
                raise ValueError('Not allow to send message to remote in standalone mode')

            return remote_func(*args, **kwargs)

    def _send_local(self, ActorRef actor_ref, object message, bint wait_response=True,
                    bint wait=True, object callback=None):
        cdef ActorExecutionContext actor_ctx
        cdef object p
        cdef object future
        cdef MessageContext message_ctx

        # send to self's actor pool
        actor_ctx = self.pool.get_actor_execution_ctx(actor_ref.uid)
        message_ctx = MessageContext(message)
        actor_ctx.receive(message_ctx)
        gevent.spawn(actor_ctx.fire_run)
        p = gevent.spawn(message_ctx.result)

        if wait:
            if wait_response:
                return p.get()
            else:
                return

        future = gevent.event.AsyncResult()

        if wait_response:
            # send, not wait
            def on_success(g):
                ret = g.value
                if callback is not None:
                    callback(ret)
                future.set_result(ret)

            def on_failure(g):
                try:
                    g.get()
                except:
                    t, ex, tb = sys.exc_info()
                    future.set_exception(ex, exc_info=(t, ex, tb))

            p.link_value(on_success)
            p.link_exception(on_failure)
        else:
            if callback is not None:
                callback(None)
            # tell, directly set future result None
            future.set_result(None)

        return future

    def _send_process(self, int to_index, ActorRef actor_ref, message, bint wait_response=True,
                      bint wait=True, callback=None):
        cdef object func
        cdef list msg
        cdef object future

        try:
            if wait_response:
                msg = pack_send_message(self.index, to_index, actor_ref, message)
            else:
                msg = pack_tell_message(self.index, to_index, actor_ref, message)
            message_id, message = msg[0], msg[1:]
        except (AttributeError, pickle.PickleError):
            raise pickle.PicklingError(f'Unable to pickle message: {message!r}')

        if wait:
            self.pipe.put(*message)
            return self.submit(message_id)

        future = gevent.event.AsyncResult()

        def check():
            self.pipe.put(*message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    cpdef object _send_remote(self, ActorRef actor_ref, object message,
                              bint wait_response=True, bint wait=True, object callback=None):
        if wait_response:
            return self.remote_handler.send(actor_ref, message, wait=wait, callback=callback)
        else:
            return self.remote_handler.tell(actor_ref, message, wait=wait, callback=callback)

    cpdef _send(self, ActorRef actor_ref, object message, bint wait_response=True,
                bint wait=True, object callback=None):
        return self._dispatch(self._send_local, self._send_process, self._send_remote, actor_ref,
                              (actor_ref, message),
                              dict(wait_response=wait_response, wait=wait, callback=callback))

    cpdef send(self, ActorRef actor_ref, object message, bint wait=True, object callback=None):
        return self._send(actor_ref, message, wait_response=True, wait=wait, callback=callback)

    cpdef tell(self, ActorRef actor_ref, object message, object delay=None,
               bint wait=True, object callback=None):
        if delay is not None:
            return gevent.spawn_later(delay, self._send, actor_ref, message, wait_response=False,
                                      wait=wait, callback=callback)

        return self._send(actor_ref, message, wait_response=False, wait=wait, callback=callback)

    cpdef object _create_local_actor(self, ActorRef actor_ref, actor_cls, args, kwargs):
        cdef bint wait
        cdef object callback
        cdef object future
        cdef object res

        wait = kwargs.pop('wait', True)
        callback = kwargs.pop('callback', None)

        if wait:
            return self.pool.create_actor(actor_cls, actor_ref.uid, *args, **kwargs)

        future = gevent.event.AsyncResult()
        try:
            res = self.pool.create_actor(actor_cls, actor_ref.uid, *args, **kwargs)
            if callback is not None:
                callback(res)
            future.set_result(res)
        except:
            t, ex, tb = sys.exc_info()
            future.set_exception(ex, exc_info=(t, ex, tb))

        return future

    def _create_process_actor(self, int to_index, ActorRef actor_ref, object actor_cls,
                              args, kwargs):
        cdef bint wait
        cdef object callback
        cdef bytes message_id
        cdef bytearray message
        cdef object future

        wait = kwargs.pop('wait', True)
        callback = kwargs.pop('callback', None)

        try:
            message_id, message = pack_create_actor_message(
                self.index, to_index, actor_ref, actor_cls, args, kwargs)
        except (AttributeError, pickle.PickleError):
            raise pickle.PicklingError(f'Unable to pickle {actor_cls}(*{args!r}, **{kwargs!r})')

        if wait:
            self.pipe.put(message)
            return self.submit(message_id)

        # return future
        future = gevent.event.AsyncResult()

        def check():
            self.pipe.put(message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    cpdef object _create_remote_actor(self, ActorRef actor_ref, object actor_cls, args, kwargs):
        return self.remote_handler.create_actor(actor_ref.address, actor_ref.uid,
                                                actor_cls, *args, **kwargs)

    def create_actor(self, str address, object uid,
                     object actor_cls, *args, **kwargs):
        cdef object actor_id
        cdef ActorRef actor_ref

        actor_id = uid or new_actor_id()
        actor_ref = ActorRef(address, actor_id)
        return self._dispatch(self._create_local_actor, self._create_process_actor,
                              self._create_remote_actor, actor_ref,
                              (actor_ref, actor_cls, args, kwargs), dict())

    cpdef _destroy_local_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        cdef object future
        cdef object res
        cdef object t
        cdef object ex
        cdef object tb

        if wait:
            return self.pool.destroy_actor(actor_ref.uid)

        future = gevent.event.AsyncResult()
        try:
            res = self.pool.destroy_actor(actor_ref.uid)
            if callback is not None:
                callback(res)
            future.set_result(res)
        except:
            t, ex, tb = sys.exc_info()
            future.set_exception(ex, exc_info=(t, ex, tb))

        return future

    def _destroy_process_actor(self, int to_index, ActorRef actor_ref,
                               bint wait=True, object callback=None):
        cdef bytes message_id
        cdef bytearray message
        cdef object future

        message_id, message = pack_destroy_actor_message(self.index, to_index, actor_ref)
        if wait:
            self.pipe.put(message)
            return self.submit(message_id)

        future = gevent.event.AsyncResult()

        def check():
            self.pipe.put(message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    cpdef destroy_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        return self._dispatch(self._destroy_local_actor, self._destroy_process_actor,
                              self.remote_handler.destroy_actor, actor_ref,
                              (actor_ref,), dict(wait=wait, callback=callback))

    cpdef _has_local_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        if wait:
            return self.pool.has_actor(actor_ref.uid)

        future = gevent.event.AsyncResult()
        try:
            ret = self.pool.has_actor(actor_ref.uid)
            if callback is not None:
                callback(ret)
            future.set_result(ret)
        except:
            t, ex, tb = sys.exc_info()
            future.set_exception(ex, exc_info=(t, ex, tb))

        return future

    def _has_process_actor(self, int to_index, ActorRef actor_ref,
                           bint wait=True, object callback=None):
        cdef bytes message_id
        cdef bytearray message
        cdef object future

        message_id, message = pack_has_actor_message(self.index, to_index, actor_ref)

        if wait:
            self.pipe.put(message)
            return self.submit(message_id)

        future = gevent.event.AsyncResult()

        def check():
            self.pipe.put(message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    cpdef object has_actor(self, ActorRef actor_ref, bint wait=True, callback=None):
        return self._dispatch(self._has_local_actor, self._has_process_actor,
                              self.remote_handler.has_actor, actor_ref,
                              (actor_ref,), dict(wait=wait, callback=callback))

    cpdef _on_receive_send(self, bytes binary, object callback):
        cdef object message
        cdef MessageContext message_ctx
        cdef ActorExecutionContext actor_ctx
        cdef bytearray result_binary
        cdef object t
        cdef object ex
        cdef object tb

        message = unpack_send_message(binary)
        try:
            message_ctx = MessageContext(message.message)
            actor_ctx = self.pool.get_actor_execution_ctx(message.actor_ref.uid)
            actor_ctx.receive(message_ctx)
            gevent.spawn(actor_ctx.fire_run)

            _, result_binary = pack_result_message(
                message.message_id, self.index, message.from_index, message_ctx.result())
        except:
            t, ex, tb = sys.exc_info()
            _, result_binary = pack_error_message(
                message.message_id, self.index, message.from_index, t, ex, tb)

        return callback(result_binary)

    cpdef _on_receive_tell(self, bytes binary, object callback):
        cdef object message
        cdef ActorExecutionContext actor_ctx
        cdef bytearray result_binary
        cdef object t
        cdef object ex
        cdef object tb

        message = unpack_tell_message(binary)
        try:
            actor_ctx = self.pool.get_actor_execution_ctx(message.actor_ref.uid)
            actor_ctx.receive(MessageContext(message.message))
            # do not wait for the completion of greenlet
            gevent.spawn(actor_ctx.fire_run)
            _, result_binary = pack_result_message(
                message.message_id, self.index, message.from_index, None)
        except:
            t, ex, tb = sys.exc_info()
            _, result_binary = pack_error_message(
                message.message_id, self.index, message.from_index, t, ex, tb)

        return callback(result_binary)

    cpdef _on_receive_create_actor(self, bytes binary, object callback):
        cdef object message
        cdef ActorExecutionContext actor_ctx
        cdef bytearray result_binary
        cdef object t
        cdef object ex
        cdef object tb

        message = unpack_create_actor_message(binary)
        try:
            _, result_binary = pack_result_message(
                message.message_id, self.index, message.from_index,
                self.pool.create_actor(message.actor_cls, message.actor_ref.uid,
                                       *message.args, **message.kwargs))
        except:
            t, ex, tb = sys.exc_info()
            _, result_binary = pack_error_message(
                message.message_id, self.index, message.from_index, t, ex, tb)

        return callback(result_binary)

    cpdef _on_receive_destroy_actor(self, bytes binary, object callback):
        cdef object message
        cdef bytearray result_binary
        cdef object t
        cdef object ex
        cdef object tb

        message = unpack_destroy_actor_message(binary)

        try:
            _, result_binary = pack_result_message(
                message.message_id, self.index, message.from_index,
                self.pool.destroy_actor(message.actor_ref.uid))
        except:
            t, ex, tb = sys.exc_info()
            _, result_binary = pack_error_message(
                message.message_id, self.index, message.from_index, t, ex, tb)

        return callback(result_binary)

    cpdef _on_receive_has_actor(self, bytes binary, object callback):
        cdef object message
        cdef bytearray result_binary
        cdef object t
        cdef object ex
        cdef object tb

        message = unpack_has_actor_message(binary)

        try:
            _, result_binary = pack_result_message(
                message.message_id, self.index, message.from_index,
                self.pool.has_actor(message.actor_ref.uid))
        except:
            t, ex, tb = sys.exc_info()
            _, result_binary = pack_error_message(
                message.message_id, self.index, message.from_index, t, ex, tb)

        return callback(result_binary)

    cpdef _on_receive_result(self, bytes binary, object callback):
        cdef object message

        message = unpack_result_message(binary)
        self.got(message.message_id, message.result)

    cpdef _on_receive_error(self, bytes binary, object callback):
        cdef object message

        message = unpack_error_message(binary)
        self.err(message.message_id, message.error_type, message.error, message.traceback)

    cpdef on_receive(self, bytes binary, object callback=None):
        cdef int message_type

        if callback is None:
            callback = self.pipe.put

        message_type = unpack_message_type_value(binary)
        return self._handlers[message_type](binary, callback)

    def run(self):
        cdef bytes message

        self.running.clear()
        if self.pipe is not None:
            while True:
                message = self.pipe.get()
                gevent.spawn(self.on_receive, message)


cdef class Dispatcher(AsyncHandler):
    """
    Only created when more than 1 process.

    What dispatcher do is redirect the message to the destination process
    according to the distributor, and then return back the result to the original process.
    """

    cdef ClusterInfo cluster_info
    cdef int index
    cdef list pipes
    cdef list pipe_checkers
    cdef object distributor  # has to be object, not Distributor
    cdef object remote_handler
    cdef object checker_group
    cdef dict handlers

    def __init__(self, ClusterInfo cluster_info, list pipes, Distributor distributor=None, parallel=None):
        super().__init__()
        AsyncHandler.__init__(self)

        self.cluster_info = cluster_info
        self.index = -1
        self.pipes = pipes  # type: list
        self.pipe_checkers = []
        self.checker_group = None
        if distributor is None:
            self.distributor = Distributor(self.cluster_info.n_process)
        else:
            self.distributor = distributor
        self.remote_handler = ActorRemoteHelper(parallel=parallel)

        self.handlers = {
            MessageType.send_all: self._on_receive_action,
            MessageType.tell_all: self._on_receive_action,
            MessageType.create_actor: self._on_receive_action,
            MessageType.destroy_actor: self._on_receive_action,
            MessageType.has_actor: self._on_receive_action,
            MessageType.result: self._on_receive_result,
            MessageType.error: self._on_receive_error,
        }

    cdef inline bint _is_remote(self, ActorRef actor_ref):
        if actor_ref.address is None or actor_ref.address == self.cluster_info.address \
                or actor_ref.address == self.cluster_info.advertise_address:
            return False
        return True

    def _send(self, ActorRef actor_ref, object message, bint wait_response=True,
                      bint wait=True, object callback=None):
        cdef int to_index
        cdef object func
        cdef list msg
        cdef bytes message_id
        cdef list messages
        cdef object future

        if self._is_remote(actor_ref):
            if wait_response:
                return self.remote_handler.send(actor_ref, message, wait=wait, callback=callback)
            else:
                return self.remote_handler.tell(actor_ref, message, wait=wait, callback=callback)

        to_index = self.distributor.distribute(actor_ref.uid)
        try:
            if wait_response:
                msg = pack_send_message(self.index, to_index, actor_ref, message)
            else:
                msg = pack_tell_message(self.index, to_index, actor_ref, message)
            message_id, messages = msg[0], msg[1:]
        except (AttributeError, pickle.PickleError):
            raise pickle.PicklingError(f'Unable to pickle message: {message!r}')

        if wait:
            self.pipes[to_index].put(*messages)
            return self.submit(message_id)

        # return future
        future = gevent.event.AsyncResult()

        def check():
            self.pipes[to_index].put(*messages)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    cpdef object send(self, ActorRef actor_ref, object message, bint wait=True, object callback=None):
        return self._send(actor_ref, message, wait_response=True, wait=wait, callback=callback)

    cpdef object tell(self, ActorRef actor_ref, object message, object delay=None,
                      bint wait=True, object callback=None):
        if delay is not None:
            return gevent.spawn_later(delay, self._send, actor_ref, message, wait_response=False,
                                      wait=wait, callback=callback)

        return self._send(actor_ref, message, wait_response=False, wait=wait, callback=callback)

    def create_actor(self, str address, object uid, object actor_cls, *args, **kwargs):
        cdef object actor_id
        cdef ActorRef actor_ref
        cdef bint wait
        cdef object callback
        cdef bytes message_id
        cdef bytearray message
        cdef int to_index
        cdef object future

        actor_id = uid or new_actor_id()
        actor_ref = ActorRef(address, actor_id)

        if self._is_remote(actor_ref):
            return self.remote_handler.create_actor(address, uid, actor_cls, *args, **kwargs)

        wait = kwargs.pop('wait', True)
        callback = kwargs.pop('callback', None)

        to_index = self.distributor.distribute(actor_id)
        try:
            message_id, message = pack_create_actor_message(
                self.index, to_index, actor_ref, actor_cls, args, kwargs)
        except (AttributeError, pickle.PickleError):
            raise pickle.PicklingError(f'Unable to pickle {actor_cls}(*{args!r}, **{kwargs!r})')

        if wait:
            self.pipes[to_index].put(message)
            return self.submit(message_id)

        # return future
        future = gevent.event.AsyncResult()

        def check():
            self.pipes[to_index].put(message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    def destroy_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        cdef int to_index
        cdef bytes message_id
        cdef bytearray message
        cdef object future

        if self._is_remote(actor_ref):
            return self.remote_handler.destroy_actor(actor_ref, wait=wait, callback=callback)

        to_index = self.distributor.distribute(actor_ref.uid)

        message_id, message = pack_destroy_actor_message(
            self.index, to_index, actor_ref)

        if wait:
            self.pipes[to_index].put(message)
            return self.submit(message_id)

        # return future
        future = gevent.event.AsyncResult()

        def check():
            self.pipes[to_index].put(message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    def has_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        cdef int to_index
        cdef bytes message_id
        cdef bytearray message
        cdef object future

        if self._is_remote(actor_ref):
            return self.remote_handler.has_actor(actor_ref, wait=wait, callback=callback)

        to_index = self.distributor.distribute(actor_ref.uid)

        message_id, message = pack_has_actor_message(
            self.index, to_index, actor_ref)

        if wait:
            self.pipes[to_index].put(message)
            return self.submit(message_id)

        # return future
        future = gevent.event.AsyncResult()

        def check():
            self.pipes[to_index].put(message)
            self.future(message_id, future, callback=callback)

        gevent.spawn(check)
        return future

    cpdef _on_receive_action(self, bytes binary):
        cdef int from_index
        cdef int to_index
        cdef bytes message_id

        from_index, to_index = get_index(binary, self.distributor.distribute)
        message_id = unpack_message_id(binary)

        if from_index == REMOTE_FROM_INDEX:
            # sent from remote, redirect to process, and wait for result
            self.pipes[to_index].put(binary)
            return self.wait(message_id)
        else:
            # sent from other process, just redirect
            self.pipes[to_index].put(binary)

    cpdef _on_receive_result(self, bytes binary):
        cdef int from_index
        cdef int to_index
        cdef bytes message_id
        cdef object result

        from_index, to_index = get_index(binary, None)
        if to_index == self.index:
            # sent from the dispatcher
            result = unpack_result_message(binary, from_index=from_index, to_index=to_index)
            self.got(result.message_id, result.result)
        elif to_index == REMOTE_FROM_INDEX:
            # sent from remote
            message_id = unpack_message_id(binary)
            self.got(message_id, binary)
        else:
            # sent from process
            self.pipes[to_index].put(binary)

    cpdef _on_receive_error(self, binary):
        cdef int to_index
        cdef object err
        cdef bytes message_id

        _, to_index = get_index(binary, None)
        if to_index == self.index:
            # sent from the dispatcher
            err = unpack_error_message(binary)
            self.err(err.message_id, err.error_type, err.error, err.traceback)
        elif to_index == REMOTE_FROM_INDEX:
            # sent from remote
            message_id = unpack_message_id(binary)
            self.got(message_id, binary)
        else:
            # sent from process
            self.pipes[to_index].put(binary)

    cpdef on_receive(self, bytes binary):
        cdef int message_type

        message_type = unpack_message_type_value(binary)
        return self.handlers[message_type](binary)

    def replace_pipe(self, idx):
        new_gl = gevent.spawn(self._check_pipe, self.pipes[idx])
        self.checker_group.start(new_gl)
        self.checker_group.discard(self.pipe_checkers[idx])
        self.pipe_checkers[idx] = new_gl

    def _check_pipe(self, pipe):
        cdef bytes message
        while True:
            try:
                message = pipe.get()
            except EOFError:
                # broken pipe
                return
            try:
                gevent.spawn(self.on_receive, message)
            except BrokenPipeError:
                return

    def run(self):
        while True:
            self.checker_group = gevent.pool.Group()
            for pipe in self.pipes:
                gl = gevent.spawn(self._check_pipe, pipe)
                self.pipe_checkers.append(gl)
                self.checker_group.start(gl)
            self.checker_group.join()
            gevent.sleep(0.05)
            if all(p is None for p in self.pipes):
                break


cdef class ActorServerHandler:
    cdef object sender
    cdef bint multi_process
    cdef object server

    def __init__(self, sender, multi_process):
        self.sender = sender
        self.multi_process = multi_process
        self.server = None

    def set_server(self, server):
        self.server = weakref.ref(server)

    def on_receive(self, bytes binary):
        if not self.multi_process:
            return self.sender.on_receive(binary, callback=lambda x: x)
        else:
            return self.sender.on_receive(binary)

    def __call__(self, sock, address):
        cdef bytes binary
        cdef object p

        server = self.server()
        if server:
            server.add_sock(sock)

        while True:
            try:
                binary = read_remote_message(sock.recv)
                p = gevent.spawn(self.on_receive, binary)
                write_remote_message(sock.sendall, p.get())
            except (gevent.socket.error, struct.error):
                break


class ActorStreamServer(gevent.server.StreamServer):
    def __init__(self, listener, handler):
        super().__init__(listener, handler)
        self._socks = []
        handler.set_server(self)

    def add_sock(self, sock):
        self._socks.append(sock)

    def close(self):
        super().close()
        [sock.close() for sock in self._socks]


cpdef object start_actor_server(ClusterInfo cluster_info, object sender):
    cdef str address
    cdef int port
    cdef bint multi_process
    cdef object s

    address, port = cluster_info.location, cluster_info.port
    if address is None or port is None:
        raise ValueError('Both address and port should be provided')
    multi_process = cluster_info.n_process > 1

    s = ActorStreamServer((address, port), ActorServerHandler(sender, multi_process))
    s.start()
    return s


def start_local_pool(int index, ClusterInfo cluster_info, object pipe=None,
                     Distributor distributor=None, object parallel=None, bint join=False,
                     object post_start=None):
    # new process will pickle the numpy RandomState, we seed the random one
    import numpy as np
    np.random.seed()

    if post_start is not None:
        post_start(index)

    # all these work in a single process
    # we start a local pool to handle messages
    # and a communicator to do the redirection of messages
    local_pool = LocalActorPool(cluster_info.advertise_address or cluster_info.address, index)
    comm = Communicator(local_pool, cluster_info, pipe, distributor, parallel)
    local_pool.set_comm(comm)
    p = gevent.spawn(comm.run)
    if join:
        p.get()
    else:
        return comm


def close_pipe(p):
    if p is None:
        return
    for _ in range(3):
        try:
            p.close()
        except gipc.GIPCClosed:
            break
        except gipc.GIPCLocked:
            continue


cdef class ActorPool:
    """
    1) If only 1 process, start local pool
    2) More than 1, start dispatcher, and start local pool in several times
    """

    cdef public ClusterInfo cluster_info
    cdef public object _dispatcher
    cpdef public Distributor distributor
    cdef bint _started
    cdef object _stopped
    cdef bint _multi_process
    cdef public object _server
    cdef object _parallel
    cdef list _stop_funcs
    cdef list _processes
    cdef list _comm_pipes
    cdef list _pool_pipes

    def __init__(self, ClusterInfo cluster_info, Distributor distributor=None, parallel=None):
        self.cluster_info = cluster_info
        self.distributor = distributor

        self._started = False
        self._stopped = gevent.event.Event()
        self._multi_process = self.cluster_info.n_process > 1
        self._dispatcher = None
        self._server = None
        self._parallel = parallel
        self._stop_funcs = []
        self._processes = []
        self._comm_pipes = []
        self._pool_pipes = []

    cdef int _check_started(self) except -1:
        if not self._started:
            raise ActorPoolNotStarted('Actor pool need to run first')
        return 0

    def create_actor(self, object actor_cls, *args, **kwargs):
        cdef bint wait
        cdef object address
        cdef object uid
        cdef ActorRef actor_ref

        self._check_started()

        wait = kwargs.pop('wait', True)

        address = kwargs.pop('address', None)
        uid = kwargs.pop('uid', None)

        if wait:
            actor_ref = self._dispatcher.create_actor(address, uid,
                                                      actor_cls, *args, **kwargs)
            if address:
                actor_ref.address = address
            actor_ref.ctx = ActorContext(self._dispatcher)
            return actor_ref

        def callback(ref):
            if address is not None:
                ref.address = address
            ref.ctx = ActorContext(self._dispatcher)

        kwargs['wait'] = False
        kwargs['callback'] = callback
        return self._dispatcher.create_actor(address, uid,
                                             actor_cls, *args, **kwargs)

    cpdef object destroy_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        self._check_started()

        return self._dispatcher.destroy_actor(actor_ref, wait=wait, callback=callback)

    cpdef object has_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        self._check_started()

        return self._dispatcher.has_actor(actor_ref, wait=wait, callback=callback)

    def actor_ref(self, *args, **kwargs):
        cdef ActorRef ref

        ref = create_actor_ref(*args, **kwargs)
        ref.ctx = ActorContext(self._dispatcher)
        return ref

    @property
    def processes(self):
        return self._processes

    @property
    def address(self):
        return self.cluster_info.address

    @property
    def advertise_address(self):
        return self.cluster_info.advertise_address

    def pre_process_start(self, idx):
        pass

    def post_process_start_parent(self, idx):
        pass

    def post_process_start_child(self, idx):
        pass

    cpdef tuple _start_process(self, idx):
        comm_pipe, pool_pipe = gipc.pipe(True, encoder=_inaction_encoder, decoder=_inaction_decoder)
        self.pre_process_start(idx)
        p = gipc.start_process(start_local_pool,
                               args=(idx, self.cluster_info, comm_pipe, self.distributor),
                               kwargs={'parallel': self._parallel, 'join': True,
                                       'post_start': self.post_process_start_child},
                               daemon=True)
        self.post_process_start_parent(idx)
        return p, comm_pipe, pool_pipe

    cpdef restart_process(self, int idx):
        if self._processes[idx].is_alive():
            self._processes[idx].terminate()
        close_pipe(self._comm_pipes[idx])
        close_pipe(self._pool_pipes[idx])
        self._processes[idx], self._comm_pipes[idx], self._pool_pipes[idx] = self._start_process(idx)
        self._dispatcher.replace_pipe(idx)

    def run(self):
        if self._started:
            return

        if not self._multi_process:
            # only start local pool
            self._dispatcher = start_local_pool(0, self.cluster_info, distributor=self.distributor,
                                                parallel=self._parallel)
        else:
            self._processes, self._comm_pipes, self._pool_pipes = [list(tp) for tp in zip(
                *(self._start_process(idx) for idx in range(self.cluster_info.n_process))
            )]

            def stop_func():
                for process in self._processes:
                    process.terminate()
                for idx, p in enumerate(self._comm_pipes):
                    close_pipe(p)
                    self._comm_pipes[idx] = None
                for idx, p in enumerate(self._pool_pipes):
                    close_pipe(p)
                    self._pool_pipes[idx] = None

            self._stop_funcs.append(stop_func)

            # start dispatcher
            self._dispatcher = Dispatcher(self.cluster_info, self._pool_pipes,
                                          self.distributor)
            gevent.spawn(self._dispatcher.run)

        if not self.cluster_info.standalone:
            # start stream server to handle remote message
            self._server = start_actor_server(self.cluster_info, self._dispatcher)

            def close():
                self._server.stop()

            self._stop_funcs.append(close)

        self._started = True

    @staticmethod
    def sleep(seconds):
        gevent.sleep(seconds)

    def join(self, timeout=None):
        self._stopped.wait(timeout)

    cpdef stop(self):
        try:
            for stop_func in self._stop_funcs:
                stop_func()
        finally:
            self._stopped.set()
            self._started = False

    def __enter__(self):
        self.run()
        return self

    def __exit__(self, *_):
        self.stop()


cdef class ActorClient:
    cdef object remote_handler

    def __init__(self, parallel=None):
        self.remote_handler = ActorRemoteHelper(parallel)

    def create_actor(self, object actor_cls, *args, **kwargs):
        cdef object address
        cdef object uid

        if 'address' not in kwargs or kwargs.get('address') is None:
            raise ValueError('address must be provided')
        address = kwargs.pop('address')
        uid = kwargs.pop('uid', new_actor_id())
        return self.remote_handler.create_actor(address, uid, actor_cls, *args, **kwargs)

    cpdef object has_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        return self.remote_handler.has_actor(actor_ref, wait=wait, callback=callback)

    cpdef destroy_actor(self, ActorRef actor_ref, bint wait=True, object callback=None):
        return self.remote_handler.destroy_actor(actor_ref, wait=wait, callback=callback)

    def actor_ref(self, *args, **kwargs):
        cdef ActorRef ref

        ref = self.remote_handler.actor_ref(*args, **kwargs)
        if ref.address is None:
            raise ValueError('address must be provided')
        return ref

    @staticmethod
    def event():
        return gevent.event.Event()

    @staticmethod
    def sleep(seconds):
        gevent.sleep(seconds)

    @staticmethod
    def popen(*args, **kwargs):
        return gevent.subprocess.Popen(*args, **kwargs)

    @staticmethod
    def threadpool(size):
        return GeventThreadPool(size)

    @staticmethod
    def asyncpool(size=None):
        return gevent.pool.Pool(size)

    @staticmethod
    def lock():
        return gevent.lock.Semaphore()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorReduction, TensorReductionMixin


class TensorSum(TensorReduction, TensorReductionMixin):
    _op_type_ = OperandDef.SUM
    _func_name = 'sum'

    def __init__(self, axis=None, dtype=None, keepdims=None, combine_size=None, stage=None, **kw):
        stage = self._rewrite_stage(stage)
        super().__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                         _combine_size=combine_size, _stage=stage, **kw)


def sum(a, axis=None, dtype=None, out=None, keepdims=None, combine_size=None):
    """
    Sum of tensor elements over a given axis.

    Parameters
    ----------
    a : array_like
        Elements to sum.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a sum is performed.  The default,
        axis=None, will sum all of the elements of the input tensor.  If
        axis is negative it counts from the last to the first axis.

        If axis is a tuple of ints, a sum is performed on all of the axes
        specified in the tuple instead of a single axis or all the axes as
        before.
    dtype : dtype, optional
        The type of the returned tensor and of the accumulator in which the
        elements are summed.  The dtype of `a` is used by default unless `a`
        has an integer dtype of less precision than the default platform
        integer.  In that case, if `a` is signed then the platform integer
        is used while if `a` is unsigned then an unsigned integer of the
        same precision as the platform integer is used.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must have
        the same shape as the expected output, but the type of the output
        values will be cast if necessary.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input tensor.

        If the default value is passed, then `keepdims` will not be
        passed through to the `sum` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    sum_along_axis : Tensor
        An array with the same shape as `a`, with the specified
        axis removed.   If `a` is a 0-d tensor, or if `axis` is None, a scalar
        is returned.  If an output array is specified, a reference to
        `out` is returned.

    See Also
    --------
    Tensor.sum : Equivalent method.

    cumsum : Cumulative sum of tensor elements.

    trapz : Integration of tensor values using the composite trapezoidal rule.

    mean, average

    Notes
    -----
    Arithmetic is modular when using integer types, and no error is
    raised on overflow.

    The sum of an empty array is the neutral element 0:

    >>> import mars.tensor as mt

    >>> mt.sum([]).execute()
    0.0

    Examples
    --------
    >>> mt.sum([0.5, 1.5]).execute()
    2.0
    >>> mt.sum([0.5, 0.7, 0.2, 1.5], dtype=mt.int32).execute()
    1
    >>> mt.sum([[0, 1], [0, 5]]).execute()
    6
    >>> mt.sum([[0, 1], [0, 5]], axis=0).execute()
    array([0, 6])
    >>> mt.sum([[0, 1], [0, 5]], axis=1).execute()
    array([1, 5])

    If the accumulator is too small, overflow occurs:

    >>> mt.ones(128, dtype=mt.int8).sum(dtype=mt.int8).execute()
    -128

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.empty((1,), dtype=a.dtype).sum().dtype
    else:
        dtype = np.dtype(dtype)
    op = TensorSum(axis=axis, dtype=dtype, keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/object.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/object.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\"mars/serialize/protos/object.proto\x1a!mars/serialize/protos/value.proto\"~\n\x0eObjectChunkDef\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x11\n\x05index\x18\x02 \x03(\rB\x02\x10\x01\x12\x12\n\x02op\x18\x03 \x01(\x0b\x32\x06.Value\x12\x0e\n\x06\x63\x61\x63hed\x18\x04 \x01(\x08\x12\x1c\n\x0c\x65xtra_params\x18\x05 \x01(\x0b\x32\x06.Value\x12\n\n\x02id\x18\x06 \x01(\t\"\x90\x01\n\tObjectDef\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\x02op\x18\x03 \x01(\x0b\x32\x06.Value\x12\x17\n\x07nsplits\x18\x05 \x01(\x0b\x32\x06.Value\x12\x1f\n\x06\x63hunks\x18\x06 \x03(\x0b\x32\x0f.ObjectChunkDef\x12\x1c\n\x0c\x65xtra_params\x18\x07 \x01(\x0b\x32\x06.Value\x12\n\n\x02id\x18\x08 \x01(\tb\x06proto3')
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,])




_OBJECTCHUNKDEF = _descriptor.Descriptor(
  name='ObjectChunkDef',
  full_name='ObjectChunkDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ObjectChunkDef.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='ObjectChunkDef.index', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='op', full_name='ObjectChunkDef.op', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cached', full_name='ObjectChunkDef.cached', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra_params', full_name='ObjectChunkDef.extra_params', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='ObjectChunkDef.id', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=199,
)


_OBJECTDEF = _descriptor.Descriptor(
  name='ObjectDef',
  full_name='ObjectDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ObjectDef.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='op', full_name='ObjectDef.op', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nsplits', full_name='ObjectDef.nsplits', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chunks', full_name='ObjectDef.chunks', index=3,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra_params', full_name='ObjectDef.extra_params', index=4,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='ObjectDef.id', index=5,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=202,
  serialized_end=346,
)

_OBJECTCHUNKDEF.fields_by_name['op'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OBJECTCHUNKDEF.fields_by_name['extra_params'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OBJECTDEF.fields_by_name['op'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OBJECTDEF.fields_by_name['nsplits'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OBJECTDEF.fields_by_name['chunks'].message_type = _OBJECTCHUNKDEF
_OBJECTDEF.fields_by_name['extra_params'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
DESCRIPTOR.message_types_by_name['ObjectChunkDef'] = _OBJECTCHUNKDEF
DESCRIPTOR.message_types_by_name['ObjectDef'] = _OBJECTDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ObjectChunkDef = _reflection.GeneratedProtocolMessageType('ObjectChunkDef', (_message.Message,), dict(
  DESCRIPTOR = _OBJECTCHUNKDEF,
  __module__ = 'mars.serialize.protos.object_pb2'
  # @@protoc_insertion_point(class_scope:ObjectChunkDef)
  ))
_sym_db.RegisterMessage(ObjectChunkDef)

ObjectDef = _reflection.GeneratedProtocolMessageType('ObjectDef', (_message.Message,), dict(
  DESCRIPTOR = _OBJECTDEF,
  __module__ = 'mars.serialize.protos.object_pb2'
  # @@protoc_insertion_point(class_scope:ObjectDef)
  ))
_sym_db.RegisterMessage(ObjectDef)


_OBJECTCHUNKDEF.fields_by_name['index']._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import ValueType, KeyField, TupleField, StringField
from ...core import ExecutableTuple
from ..operands import TensorHasInput, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from ..core import TensorOrder


class TensorUnravelIndex(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.UNRAVEL_INDEX

    _input = KeyField('input')
    _dims = TupleField('dims', ValueType.int32)
    _order = StringField('order')

    def __init__(self, dims=None, dtype=None, order=None, **kw):
        super().__init__(_dims=dims, _dtype=dtype, _order=order, **kw)
        if self._order is None:
            self._order = 'C'

    @property
    def dims(self):
        return self._dims

    @property
    def order(self):
        return self._order

    @property
    def output_limit(self):
        return float('inf')

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, indices):
        order = TensorOrder.C_ORDER if self._order == 'C' else TensorOrder.F_ORDER
        kws = [{'pos': i, 'order': order} for i in range(len(self._dims))]
        return ExecutableTuple(self.new_tensors([indices], indices.shape, kws=kws, output_limit=len(kws)))

    @classmethod
    def tile(cls, op):
        indices = op.inputs[0]
        dims = op.dims
        order = op.outputs[0].order

        out_chunks = [list() for _ in range(len(dims))]
        for in_chunk in indices.chunks:
            chunk_op = op.copy().reset_key()
            chunk_kws = [{'pos': i, 'index': in_chunk.index, 'order': order}
                         for i in range(len(dims))]
            chunks = chunk_op.new_chunks([in_chunk], shape=in_chunk.shape, kws=chunk_kws,
                                         output_limit=len(dims))
            for out_chunk, c in zip(out_chunks, chunks):
                out_chunk.append(c)

        new_op = op.copy()
        kws = [{'chunks': out_chunk, 'nsplits': indices.nsplits, 'shape': o.shape}
               for out_chunk, o in zip(out_chunks, op.outputs)]
        return new_op.new_tensors(op.inputs, kws=kws, output_limit=len(dims), order=order)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)
        indices = inputs[0]

        with device(device_id):
            outputs = xp.unravel_index(indices, op.dims, order=op.order)
            for o, output in zip(op.outputs, outputs):
                ctx[o.key] = output


def unravel_index(indices, dims, order='C'):
    """
    Converts a flat index or tensor of flat indices into a tuple
    of coordinate tensors.

    Parameters
    ----------
    indices : array_like
        An integer tensor whose elements are indices into the flattened
        version of a tensor of dimensions ``dims``.
    dims : tuple of ints
        The shape of the tensor to use for unraveling ``indices``.
    order : {'C', 'F'}, optional
        Determines whether the indices should be viewed as indexing in
        row-major (C-style) or column-major (Fortran-style) order.

    Returns
    -------
    unraveled_coords : tuple of Tensor
        Each tensor in the tuple has the same shape as the ``indices``
        tensor.

    See Also
    --------
    ravel_multi_index

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.unravel_index([22, 41, 37], (7,6)).execute()
    (array([3, 6, 6]), array([4, 5, 1]))

    >>> mt.unravel_index(1621, (6,7,8,9)).execute()
    (3, 1, 4, 1)
    """
    indices = astensor(indices)
    if isinstance(dims, Iterable):
        dims = tuple(dims)
    else:
        dims = (dims,)

    if order not in 'CF':
        raise TypeError("only 'C' or 'F' order is permitted")

    op = TensorUnravelIndex(dims=dims, dtype=np.dtype(np.intp), order=order)
    return op(indices)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import AnyField
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorHypergeometric(TensorDistribution, TensorRandomOperandMixin):
    __slots__ = '_ngood', '_nbad', '_nsample', '_size'
    _input_fields_ = ['_ngood', '_nbad', '_nsample']
    _op_type_ = OperandDef.RAND_HYPERGEOMETRIC

    _ngood = AnyField('ngood')
    _nbad = AnyField('nbad')
    _nsample = AnyField('nsample')
    _func_name = 'hypergeometric'

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super().__init__(_state=state, _size=size, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def ngood(self):
        return self._ngood

    @property
    def nbad(self):
        return self._nbad

    @property
    def nsample(self):
        return self._nsample

    def __call__(self, ngood, nbad, nsample, chunk_size=None):
        return self.new_tensor([ngood, nbad, nsample], None, raw_chunk_size=chunk_size)


def hypergeometric(random_state, ngood, nbad, nsample, size=None, chunk_size=None, gpu=None, dtype=None):
    r"""
    Draw samples from a Hypergeometric distribution.

    Samples are drawn from a hypergeometric distribution with specified
    parameters, ngood (ways to make a good selection), nbad (ways to make
    a bad selection), and nsample = number of items sampled, which is less
    than or equal to the sum ngood + nbad.

    Parameters
    ----------
    ngood : int or array_like of ints
        Number of ways to make a good selection.  Must be nonnegative.
    nbad : int or array_like of ints
        Number of ways to make a bad selection.  Must be nonnegative.
    nsample : int or array_like of ints
        Number of items sampled.  Must be at least 1 and at most
        ``ngood + nbad``.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``ngood``, ``nbad``, and ``nsample``
        are all scalars.  Otherwise, ``np.broadcast(ngood, nbad, nsample).size``
        samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized hypergeometric distribution.

    See Also
    --------
    scipy.stats.hypergeom : probability density function, distribution or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the Hypergeometric distribution is

    .. math:: P(x) = \frac{\binom{m}{n}\binom{N-m}{n-x}}{\binom{N}{n}},

    where :math:`0 \le x \le m` and :math:`n+m-N \le x \le n`

    for P(x) the probability of x successes, n = ngood, m = nbad, and
    N = number of samples.

    Consider an urn with black and white marbles in it, ngood of them
    black and nbad are white. If you draw nsample balls without
    replacement, then the hypergeometric distribution describes the
    distribution of black balls in the drawn sample.

    Note that this distribution is very similar to the binomial
    distribution, except that in this case, samples are drawn without
    replacement, whereas in the Binomial case samples are drawn with
    replacement (or the sample space is infinite). As the sample space
    becomes large, this distribution approaches the binomial.

    References
    ----------
    .. [1] Lentner, Marvin, "Elementary Applied Statistics", Bogden
           and Quigley, 1972.
    .. [2] Weisstein, Eric W. "Hypergeometric Distribution." From
           MathWorld--A Wolfram Web Resource.
           http://mathworld.wolfram.com/HypergeometricDistribution.html
    .. [3] Wikipedia, "Hypergeometric distribution",
           http://en.wikipedia.org/wiki/Hypergeometric_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> ngood, nbad, nsamp = 100, 2, 10
    # number of good, number of bad, and number of samples
    >>> s = mt.random.hypergeometric(ngood, nbad, nsamp, 1000)
    >>> hist(s)
    #   note that it is very unlikely to grab both bad items

    Suppose you have an urn with 15 white and 15 black marbles.
    If you pull 15 marbles at random, how likely is it that
    12 or more of them are one color?

    >>> s = mt.random.hypergeometric(15, 15, 15, 100000)
    >>> (mt.sum(s>=12)/100000. + mt.sum(s<=3)/100000.).execute()
    #   answer = 0.003 ... pretty unlikely!
    """
    if dtype is None:
        dtype = np.random.RandomState().hypergeometric(
            handle_array(ngood), handle_array(nbad), handle_array(nsample), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorHypergeometric(state=random_state.to_numpy(), size=size, gpu=gpu, dtype=dtype)
    return op(ngood, nbad, nsample, chunk_size=chunk_size)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/graph.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import anyref_pb2 as mars_dot_serialize_dot_protos_dot_anyref__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/graph.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n!mars/serialize/protos/graph.proto\x1a\"mars/serialize/protos/anyref.proto\"g\n\x08GraphDef\x12\x1e\n\x05level\x18\x02 \x01(\x0e\x32\x0f.GraphDef.Level\x12\x1b\n\x04node\x18\x01 \x03(\x0b\x32\r.AnyReference\"\x1e\n\x05Level\x12\t\n\x05\x43HUNK\x10\x00\x12\n\n\x06\x45NTITY\x10\x01\x62\x06proto3')
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_anyref__pb2.DESCRIPTOR,])



_GRAPHDEF_LEVEL = _descriptor.EnumDescriptor(
  name='Level',
  full_name='GraphDef.Level',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CHUNK', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENTITY', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=146,
  serialized_end=176,
)
_sym_db.RegisterEnumDescriptor(_GRAPHDEF_LEVEL)


_GRAPHDEF = _descriptor.Descriptor(
  name='GraphDef',
  full_name='GraphDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='level', full_name='GraphDef.level', index=0,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node', full_name='GraphDef.node', index=1,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GRAPHDEF_LEVEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=176,
)

_GRAPHDEF.fields_by_name['level'].enum_type = _GRAPHDEF_LEVEL
_GRAPHDEF.fields_by_name['node'].message_type = mars_dot_serialize_dot_protos_dot_anyref__pb2._ANYREFERENCE
_GRAPHDEF_LEVEL.containing_type = _GRAPHDEF
DESCRIPTOR.message_types_by_name['GraphDef'] = _GRAPHDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GraphDef = _reflection.GeneratedProtocolMessageType('GraphDef', (_message.Message,), dict(
  DESCRIPTOR = _GRAPHDEF,
  __module__ = 'mars.serialize.protos.graph_pb2'
  # @@protoc_insertion_point(class_scope:GraphDef)
  ))
_sym_db.RegisterMessage(GraphDef)


# @@protoc_insertion_point(module_scope)
